/**
 * Accumulator.java
 *
 * Zilog Z80 CPU Accumulator Register Emulation
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public final class Accumulator
{
	private static int reg;
	private static int reg2;
	private static Flag f;

	/**
	 *	Accumulator Constructor
	 *
	 *	@param	fl	Pointer to Flag Register
     */

	public Accumulator(Flag fl)
	{
		reg  = 0;
		reg2 = 0;
		this.f = fl;
	}

	/**
	 *  Clear Register
	 */

	public void clear()
	{
		reg  = 0;
		reg2 = 0;
	}

	/**
	 *	Get Value of Accumulator
	 *
	 *	@return				Value Stored in Register
     */

	public int get()
	{
		return reg;
	}

	/**
	 *	Set Register
	 *
	 *	@param value			Value to set Register (0 - 0xFF)
     */

	public void set(int value)
	{
		reg = value&0xff;
	}

	/**
	 *	Increment Register
     */

	public void inc()
	{
		reg = (reg + 1)&0xff;

		if ((reg & 0x0F) == 0x00)
			f.halfCarryOn();
		else
			f.halfCarryOff();

		f.negativeOff();

		// Overflow has occurred (to last bit)
		if (reg == 0x80) f.parityOn();
		else f.parityOff();

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();
	}

	/**
	 *	Decrement Register
     */

	public void dec()
	{
		reg = (reg - 1)&0xff;

		if ((reg & 0x0F) == 0x0F)
			f.halfCarryOn();
		else
			f.halfCarryOff();

		f.negativeOn();

		// Overflow has occurred (from last bit)
		if (reg == 0x7F) f.parityOn();
		else f.parityOff();

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();
	}

	/**
	 * Exchange Register Banks
	 */

	public void ex()
	{
		int temp;
		temp = reg; reg = reg2; reg2 = temp;
	}

	/**
	 *	ADD 8 BIT
	 *
	 *  @param value		Value to add
     */

	public void add(int value)
	{
		int result = reg + value;

		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x10) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x100) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) f.parityOn();
	    else f.parityOff();

		reg = result&0xFF;

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		f.negativeOff();

		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	ADC 8 BIT - Add with carry
	 *
	 *  @param value		Value to add
     */

	public void adc(int value)
	{
		int result = reg + value + (f.carry() ? 1 : 0);
		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x10) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x100) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) f.parityOn();
	    else f.parityOff();

		reg = result&0xFF;

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		f.negativeOff();
		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	SUB 8 BIT
	 *
	 *  @param value		Value to subtract
     */

	public void sub(int value)
	{
		int result = reg - value;
		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x10) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x100) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) f.parityOn();
	    else f.parityOff();

		reg = result&0xFF;

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		f.negativeOn();

		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	SBC 8 BIT
	 *
	 *  @param value		Subtract with carry
     */

	public void sbc(int value)
	{
		int result = reg - value - (f.carry() ? 1 : 0);
		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x10) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x100) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) f.parityOn();
	    else f.parityOff();

		reg = result&0xFF;

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		f.negativeOn();
		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	AND Operation
	 *
	 *  @param value		Value to &
     */

	public void and(int value)
	{
		// AND the value from memory
		reg &= value;

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		f.halfCarryOn();
		f.setParity(reg);
		f.negativeOff();
		f.carryOff();
		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	OR Operation
	 *
	 *  @param value		Value to |
     */

	public void or(int value)
	{
		reg |= value; // bitwise inclusive OR to turn relevant bits on

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		f.halfCarryOff();
		f.setParity(reg);
		f.negativeOff();
		f.carryOff();
		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	XOR Operation (Bitwise Exclusive OR)
	 *
	 *  @param value		Value to ^
     */

	public void xor(int value)
	{
		reg ^= value; // bitwise exclusive OR

		if (reg == 0) f.zeroOn();
		else f.zeroOff();

		if ((reg & 0x80) == 0) f.signOff();
		else f.signOn();

		f.halfCarryOff();
		f.negativeOff();
		f.carryOff();
		f.setParity(reg);
		f.setBit3(reg);
		f.setBit5(reg);
	}


	/**
	 *	CP Operation - Compare with Accumulator
	 *
	 *  @param value		Value to compare
     */

	public void cp(int value)
	{
		int result = reg - value;
			
		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x10) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x100) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x100) != 0) f.parityOn();
	    else f.parityOff();
	    
	    result &= 0xFF;
	    
	    // sign
		if ((result & 0x80) == 0) f.signOff();
		else f.signOn();

		// zero
		if (result == 0) f.zeroOn();
		else f.zeroOff();

		// negative
		f.negativeOn();

		// YF and XF flags copied from operand, not result
		f.setBit3(value);
		f.setBit5(value);
	}

	/**
	 *	CPL Operation - Complement Accumulator
     */

	public void cpl()
	{
		reg =(~reg)&0xff;
		f.negativeOn();
		f.halfCarryOn();
		f.setBit3(reg);
		f.setBit5(reg);
	}

	/**
	 *	RRA Operation - Rotate Right Accumulator
     */

	public void rra()
	{
		int temp = reg;

		// Shift Right One Bit Position
		reg >>= 1;

		// Move contents of carry flag to bit 7
		if (f.carry()) reg |= 0x80;
		//else reg &=~0x80;

		// Move original contents to carry flag
		if ((temp & 0x01) != 0) f.carryOn();
		else f.carryOff();

		f.negativeOff(); f.halfCarryOff();
	}

	/**
	 *	RLA Operation - Rotate Left Accumulator
     */

	public void rla()
	{
		int temp = reg;
		// Shift Left One Bit Position
		reg = (reg << 1)&0xff;

		// Move contents of carry flag to bit 0
		if (f.carry()) reg |= 0x01;

		// Move original contents to carry flag
		if ((temp & 0x80) == 0x80) f.carryOn();
		else f.carryOff();

		f.negativeOff(); f.halfCarryOff();
	}

	/**
	 *	RLCA Operation - Rotate Left With Carry Accumulator
     */

	public void rlca()
	{
		// Backup copy of register
		int temp = reg;

		// Shift register left
		reg = (reg << 1)&0xff;

		// Transfer Original Bit 7 to Bit 0 and Carry Flag
		if ((temp & 0x80) == 0x80)
		{
			f.carryOn();
			reg |= 0x01;
		}
		else
		{
			f.carryOff();
		}

		f.negativeOff(); f.halfCarryOff();
	}


	/**
	 *	RRCA Operation - Rotate Right With Carry Accumulator
     */

	public void rrca()
	{
		int temp = reg;
		reg >>= 1;
		if ((temp & 0x01) != 0)
		{
			f.carryOn();
			reg |= 0x80;
		}
		else
		{
			f.carryOff();
		}
		f.negativeOff(); f.halfCarryOff();
	}
}