/**
 * EmulateLoop.java
 *
 * Ties JavaGear's components together.
 * It loops the emulation with the appropriate NTSC/PAL Timing.
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

/*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public class EmulateLoop
	implements Runnable
{
	protected Vdp vdp;								// Pointer to VDP
	protected Z80 z80;								// Pointer to Z80
	protected SN76496 sn76496;						// Pointer to SN76496 PSG
	protected Memory mem;								// Pointer to Memory
	protected Screen screenpanel;						// Pointer to Java Screen Panel
	protected Thread thisThread;						// Pointer to this thread.
	
	protected boolean running;						// Stores whether thread is running
	protected int fps;								// Stores Desired FPS
	protected int no_of_scanlines;					// Number of Scanlines
	protected int cycles_per_line;					// CPU Cycles Per Scanline

	/**
	 *	EmulateLoop Constructor
	 *
	 *	@param z	Z80
	 *	@param m	Memory
	 *  @param v	Vdp
     *  @param s	SN76496
     *  @param f	FrameCount
     *	@param sp	ScreenPanel
     */

	public EmulateLoop(Z80 z, Memory m, Vdp v, SN76496 s, Screen sp)
	{
		this.mem = m;
		this.vdp = v;
		this.z80 = z;
		this.sn76496 = s;
		this.screenpanel = sp;

		// Not Running
		running = false;
		// Default to NTSC Timing
		setNTSC();
	}

	/**
	 *	Quickly Reset the Z80, Memory, VDP and Sound.
     */

	public void reset()
	{
		mem.reset();
		z80.reset();
		vdp.reset();
		sn76496.reset();
	}

	/**
	 *	Start Emulation
     */

	public synchronized void startThread()
	{
		if (thisThread == null)
		{
			running = true;
			reset();
			thisThread = new Thread(this);
			thisThread.start();
		}
	}

	/**
	 *	Stop Emulation
     */

	public synchronized void stopThread()
	{
		if (thisThread != null)
		{
			thisThread.interrupt();
			running = false;
		}
		
		while (thisThread != null);
	}

	/**
	 *	Suspend Emulation
     */

	public synchronized void suspendThread()
	{
		if (thisThread != null)
		{
			running = false;
		}
		
		while (thisThread != null);
	}

	/**
	 *	Resume suspended Emulation
     */

	public synchronized void resumeThread()
	{
		if (thisThread == null)
		{
			running = true;
			thisThread = new Thread(this);
			thisThread.start();
		}
	}

	/**
	 *	Set NTSC Timing
     */

	public void setNTSC()
	{
		fps = 60;
		no_of_scanlines = 262;
		cycles_per_line = 227;
		sn76496.setFPS(fps);
		vdp.setNTSC(true);
	}


	/**
	 *	Set PAL Timing
     */

	public void setPAL()
	{
		fps = 50;
		no_of_scanlines = 312;
		cycles_per_line = 229;
		sn76496.setFPS(fps);
		vdp.setNTSC(false);
	}

	/**
	 *	Run Emulation. This method should not be called directly.
	 *	Instead use startThread();
     */

	public void run()
	{
		if (thisThread == null) return;
		
		Throttle.init(fps, thisThread);

		while(running)
		{		
			boolean draw_frame = !Throttle.skipFrame();

			// Draw one frame
			for (int lineno = 0; lineno < no_of_scanlines; lineno++)
			{
				// Assert Interrupt Line if Necessary
				vdp.interrupts(lineno);
				// Draw Next Line
				if (draw_frame)
					vdp.drawLine(lineno);
				// Run Z80
				z80.run(cycles_per_line);
			}

			// Refresh Java Screenpanel
			if (draw_frame)
				screenpanel.refresh();
				
			// Output Sound
			sn76496.output();

			// Only Check for Pause Button once per frame to increase emulation speed
			if (Setup.pause_button)
			{
				z80.nmi();
				Setup.pause_button = false;
			}

			Throttle.throttle();
			
		} // End of Running Loop
		
		thisThread = null;

		// Loop has finished
	}
}