/**
 * Flag.java
 *
 * Zilog Z80 CPU Flag Emulation
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public final class Flag
{
	private static int f;							// Register (and second bank)
	private static int f2;

	private final static int F_CARRY     = 0x01; 	// carry (set when a standard carry occurred)
	private final static int F_NEGATIVE  = 0x02; 	// negative (set when instruction is subtraction, clear when addition)
	private final static int F_PARITY    = 0x04; 	// true indicates even parity in the result, false for 2s complement sign overflow
	private final static int F_BIT3      = 0x08; 	// bit3 (usually a copy of bit 3 of the result)
	private final static int F_HALFCARRY = 0x10; 	// half carry (set when a carry occured between bit 3 / 4 of result - used for BCD
	private final static int F_BIT5      = 0x20; 	// bit5 (usually a copy of bit 5 of the result)
	private final static int F_ZERO      = 0x40; 	// zero (set when a result is zero)
	private final static int F_SIGN      = 0x80; 	// sign (set when a result is negative)

	/**
	 *	Flag Constructor.
     */

	public Flag()
	{
		f = 0; f2 = 0;
	}


	/**
	 *	Get Value of Flag Register
	 *
	 *	@return				Value Stored in Flag Register
     */

	public int get() {return f;}


	/**
	 *	Set Flag Register
	 *
	 *	@param value		Value to set Register (0 - 0xFF)
     */

	public void set(int value) {f = value;}


	/**
	 *  Clear Register
	 */

	public void clear() {f = 0; f2 = 0;}

	/**
	 *  Carry Flag On
	 */

	public void carryOn() {f |= F_CARRY;}

	/**
	 *  Carry Flag Off
	 */

	public void carryOff() {f &=~ F_CARRY;}

	/**
	 *	Is Carry Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean carry()
	{
		return ((f & F_CARRY) == F_CARRY);
	}


	/**
	 *  Negative Flag On
	 */

	public void negativeOn() {f |= F_NEGATIVE;}


	/**
	 *  Negative Flag Off
	 */

	public void negativeOff() {f &=~ F_NEGATIVE;}

	/**
	 *	Is Negative Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean negative()
	{
		return ((f & F_NEGATIVE) == F_NEGATIVE);
	}

	/**
	 *  Parity Flag On
	 */

	public void parityOn() {f |= F_PARITY;}

	/**
	 *  Parity Flag Off
	 */

	public void parityOff() {f &=~ F_PARITY;}

	/**
	 *	Set Parity
	 *
	 *	@param v			Value to test (0 - 0xFF)
     */

	public void setParity(int n)
	{
		if (PARITY[n&0xff]) f |= F_PARITY;
		else f &=~ F_PARITY;
	}

	/**
	 *	Is Parity Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean parity()
	{
		return ((f & F_PARITY) == F_PARITY);
	}

	public void setBit3(int v)
	{
		if ((v & F_BIT3) == F_BIT3)
			f |= F_BIT3;
		else
			f &=~ F_BIT3;
	}

	public void bit3On() {f |= F_BIT3;}
	public void bit3Off() {f &=~ F_BIT3;}

	/**
	 *	Is Bit 3 Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean bit3()
	{
		return ((f & F_BIT3) == F_BIT3);
	}


	public void halfCarryOn() {f |= F_HALFCARRY;}
	public void halfCarryOff() {f &=~ F_HALFCARRY;}

	/**
	 *	Is Halfcarry Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean halfcarry()
	{
		return ((f & F_HALFCARRY) == F_HALFCARRY);
	}

	public void setBit5(int v)
	{
		if ((v & F_BIT5) == F_BIT5)
			f |= F_BIT5;
		else
			f &=~ F_BIT5;
	}

	/**
	 *  Bit 5 Flag On
	 */

	public void bit5On() {f |= F_BIT5;}

	/**
	 *  Bit 5 Flag Off
	 */

	public void bit5Off() {f &=~ F_BIT5;}

	/**
	 *	Is Bit 5 Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean bit5()
	{
		return ((f & F_BIT5) == F_BIT5);
	}

	/**
	 *  Zero Flag On
	 */

	public void zeroOn() {f |= F_ZERO;}

	/**
	 *  Zero Flag Off
	 */

	public void zeroOff() {f &=~ F_ZERO;}

	/**
	 *	Is Zero Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean zero()
	{
		return ((f & F_ZERO) == F_ZERO);
	}

	/**
	 *  Sign Flag On
	 */

	public void signOn() {f |= F_SIGN;}

	/**
	 *  Sign Flag Off
	 */

	public void signOff() {f &=~ F_SIGN;}

	/**
	 *	Is Sign Flag Set?
	 *
	 *	@return		True if set
     */

	public boolean sign()
	{
		return ((f & F_SIGN) == F_SIGN);
	}

	/**
	 * Exchange Register Banks
	 */

	public void ex()
	{
		int temp;
		temp = f; f = f2; f2 = temp;
	}

	// Pre-calculated Flag Settings
	private final static boolean PARITY[] =
	{
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		false, true, true, false, true, false, false, true, true, false, false, true, false, true, true, false,
		true, false, false, true, false, true, true, false, false, true, true, false, true, false, false, true
 	};

}