/**
 * JavaGearApplet.java
 *
 * JavaGear Applet User Interface
 *
 * @author Copyright (C) 2002 Chris White
 * @version 16th March 2002
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;
import java.net.*;			// URL Handling

public final class JavaGearApplet extends JApplet
    implements ActionListener
{
	private Memory mem;							// Memory and Paging Emulation
	private SN76496 sn76496;						// SN76596 Emulation
	private Ports port;							// Port Emulation
	private Z80 z80cpu;							// Z80 CPU Emulation
	private Controllers controllers;				// Controller Emulation
	private EmulateLoop emulateLoop = null;		// Main Emulation Loop
	private Vdp vdp;								// VDP Emulation
	private Screen screenpanel;					// Java Screen Display

	private KeyFrame keyFrame;						// Keyboard configuration Frame
	private JToolBar bar;							// Pointer to ToolBar

	private Setup setup;							// Pointer to General Parameters

	private String rom_directory = "./roms/";		// Directory containing ROM images
	private String rom_index = "index.txt";		// File containing ROM information
	private String[] filenames; 		  			// Store the filenames stored in the index file.
	private boolean[] tv_type;			  		// Store NTSC / PAL Settings for file
	private List fileList = new List(8, false);	// File Descriptions

													// Custom ROM Loader Dialog Box
	private JFrame openFrame =
		new JFrame("Open");
	private JButton okButton =
		new JButton("OK");
    private JButton cancelButton =
    	new JButton("Cancel");

    private boolean cart_loaded = false;			// Has a Cartridge Been Loaded
	private Image gameGear, splash;				// Images

	public void init()
	{
		// Load Images
		URL imageURL = JavaGearApplet.class.getResource("images/cart.gif");
		final Image cart = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/sound.gif");
		final Image soundOn = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/soundoff.gif");
		final Image soundOff = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/controller.gif");
		final Image control = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/fs0.gif");
		final Image fs0 = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/fs1.gif");
		final Image fs1 = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/fs2.gif");
		final Image fs2 = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/gamegear.jpg");
		gameGear = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGearApplet.class.getResource("images/splash.jpg");
		splash = Toolkit.getDefaultToolkit().getImage(imageURL);

		// Wait for images to load
		MediaTracker tracker = new MediaTracker(this);
		tracker.addImage(cart, 0);
		tracker.addImage(gameGear, 1);
		tracker.addImage(soundOn, 2);
		tracker.addImage(soundOff, 3);
		tracker.addImage(control, 4);
		tracker.addImage(fs0, 5);
		tracker.addImage(fs1, 6);
		tracker.addImage(fs2, 7);
		tracker.addImage(splash, 8);
		try {tracker.waitForAll();}
		catch(InterruptedException e){}

		// Create pointer to general options
		setup = new Setup();

		// Initialise Objects
		controllers 		= new Controllers(setup);
		InterruptLine irq 	= new InterruptLine();
		mem  				= new Memory(setup);
		vdp 				= new Vdp(setup, irq);
		screenpanel 		= new Screen(setup, vdp);
		sn76496 			= new SN76496(111860.78125, 11025);
		port 				= new Ports(vdp, controllers, sn76496);
		z80cpu 				= new Z80(mem, port, irq);
		emulateLoop 		= new EmulateLoop(z80cpu, mem, vdp, sn76496, screenpanel);

		// Create ToolBar
		bar = new JToolBar();

		Action openCartAction = new AbstractAction("Open Cartridge", new ImageIcon(cart))
								{ 	public void actionPerformed(ActionEvent event)
									{
										if (cart_loaded) emulateLoop.suspendThread();
										openFrame.show();
									}
								};
		Action soundAction = new AbstractAction("Toggle Sound", new ImageIcon(soundOn))
								{ 	public void actionPerformed(ActionEvent event)
									{
										sn76496.setEnabled();
										if (sn76496.isEnabled())
											// Change Icon
											putValue(Action.SMALL_ICON, new ImageIcon(soundOn));
										else
											putValue(Action.SMALL_ICON, new ImageIcon(soundOff));
										requestFocus();
									}
								};
		Action controlAction = new AbstractAction("Controller Config", new ImageIcon(control))
								{ 	public void actionPerformed(ActionEvent event)
									{
										if (keyFrame == null) keyFrame = new KeyFrame(controllers);
										keyFrame.show();
									}
								};

		Action frameSkipAction = new AbstractAction("FrameSkip", new ImageIcon(fs0))
								{ 	public void actionPerformed(ActionEvent event)
									{
										int frameskip = Throttle.fskip;

										if (++frameskip > 2) frameskip = 0;
										Throttle.setFrameSkip(frameskip);

										// Cycle Icons
										if (frameskip == 0)
											putValue(Action.SMALL_ICON, new ImageIcon(fs0));
										else if (frameskip == 1)
											putValue(Action.SMALL_ICON, new ImageIcon(fs1));
										else if (frameskip == 2)
											putValue(Action.SMALL_ICON, new ImageIcon(fs2));

										requestFocus();
									}
								};

		// Add the Actions to the ToolBar
		bar.add(openCartAction); bar.add(controlAction); bar.add(soundAction); bar.add(frameSkipAction);
		// So the toolbar doesn't take our left/right keys!
		bar.setActionMap(null);
		// Disable toolbar dragging
		bar.setFloatable(false);
		// Add ToolBar To Top Of Frame
		getContentPane().add(bar, "North");

		// Set Emulation Defaults
		port.setEurope(true); // Default to a domestic system
		setup.setSMS(); // Default to SMS mode

		// Load ROM List
		readIndexFile();

		// Add Keyboard Support
		addKeyListener(controllers);

		// Add Display Window
		getContentPane().add(screenpanel);

		// Receive Keystrokes Immediately
		requestFocus();
	}


	/**
	 *	Invoked when a button from the ROM Selector is clicked
     */

	public void actionPerformed(ActionEvent evt)
	{
		// Load ROM from List
		if (evt.getSource() == okButton)
		{
			if (fileList.getSelectedIndex() != -1)
			{
				try
				{
					String path = rom_directory+filenames[fileList.getSelectedIndex()];

					// Set PAL/NTSC
					if (tv_type[fileList.getSelectedIndex()])
						emulateLoop.setNTSC();
					else
						emulateLoop.setPAL();

					openFrame.hide();
					openCart(new URL(getCodeBase(), path));
				}
				catch(MalformedURLException e){}
			}
			requestFocus();
		}

		else if (evt.getSource() == cancelButton)
		{
			openFrame.hide();
			if (cart_loaded) 
			{
				screenpanel.disposeGraphics();
				emulateLoop.resumeThread();
			}
			requestFocus();
		}
	}


	/**
	 *	Paint Splash Screen and Game Gear Image
	 *
	 *  @param g	The graphics context to use for painting
     */

	public void paint(Graphics g)
	{
		super.paint(g);

		if (cart_loaded)
		{
			if (setup.is_gg)
				screenpanel.getGraphics().drawImage(gameGear, 0, 0, this);
		}
		else
		{
			Dimension d = screenpanel.getSize();
			screenpanel.getGraphics().drawImage(splash, 0, 0, d.width, d.height, this);
		}
	}

	/**
	 *	Read Index File From Server.
	 *	The Index File Contains The List of ROMs
     */

	public void readIndexFile()
	{
		InputStreamReader is;
		URL url;
		int linecount = 0; // stores number of lines in files

		try
		{
			url = new URL(getCodeBase(), rom_directory + rom_index);
			is = new InputStreamReader(url.openStream());
		}

		catch(Exception e)
		{
			System.out.println("\nProblem opening URL of "+rom_index);
			System.out.println(e.getMessage());
			return;
		}

		BufferedReader bis = new BufferedReader(is);

		// Count lines in file
		try
		{
			while (bis.readLine() != null) linecount++;
			// reset to start of stream
			bis = new BufferedReader(new InputStreamReader(url.openStream()));
		}
		catch (IOException e)
		{
			System.out.println("Error reading from "+rom_index);
			return;
		}

		// Create an array of the correct size
		filenames = new String[linecount];
		tv_type = new boolean[linecount];
		linecount = 0; // Reset number of lines
		String s;

		try
		{
			while ((s = bis.readLine()) != null)
			{
				// Split string at ':'
				for (int x=0; x < s.length(); x++)
				{
					if (s.charAt(x) == ':')
					{
						int filename_start = x+1;

						// Not over bounds
						if ((x+2) < s.length())
						{
							// Check for PAL
							if ((s.charAt(x+1) == 'p') && (s.charAt(x+2) == ':'))
							{
								filename_start += 2;
								tv_type[linecount] = false;
							}
							else
								tv_type[linecount] = true;
						}

						// Add file description to Combo list box
						fileList.add(s.substring(0, x), linecount); // Add at index specified by linecount
						// Add filename to the array
						filenames[linecount++] = s.substring(filename_start);
						break;
					}
				}
			} // end while
		}
		catch (IOException e)
		{
			System.out.println("Error reading from "+rom_index);
			return;
		}

		JPanel p2 = new JPanel();
		p2.add(okButton);
		p2.add(cancelButton);
		okButton.addActionListener(this);
		cancelButton.addActionListener(this);
		openFrame.getContentPane().add(fileList, "North");
		openFrame.getContentPane().add(p2, "South");
		openFrame.setResizable(false);
		openFrame.pack();
	}

	/**
	 *	Open Cartridge and Start Emulation
	 *
	 *  @param URL	URL of cartridge image
     */

	public void openCart(URL filename)
	{
		mem.reset();
		if (mem.readCart(filename))
		{
			cart_loaded = true;
			emulateLoop.stopThread();
			if (setup.is_gg)
				emulateLoop.setNTSC();
			z80cpu.reset();
			vdp.reset();
			repaint(); // repaint this window
			emulateLoop.startThread();
		}
		else
		{
			cart_loaded = false;
			repaint();
		}
	}

	/**
	 *	Stop Applet
	 */

	public void stop()
	{
		//emulateLoop.stopThread();
	}
}
