/**
 * JavaGearFrame.java
 *
 * JavaGear Application User Interface
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.io.*;			// File Input / Output
import java.net.*;			// URL Handling

public final class JavaGearFrame extends JFrame
    implements ActionListener
{
	private Memory mem;						// Memory and Paging Emulation
	private SN76496 sn76496;					// SN76596 Emulation
	private Ports port;						// Port Emulation
	private Z80 z80cpu;						// Z80 CPU Emulation
	private Controllers controllers;			// Controller Emulation
	private EmulateLoop emulateLoop;			// Main Emulation Loop
	private Vdp vdp;							// VDP Emulation
	private Screen screenpanel;				// Java Screen Display

	private boolean show_fps;					// Show FPS Counter

	private KeyFrame keyFrame;					// Keyboard configuration Frame
	private JFrame debugger;					// Debugger Frame
	private VdpDebugFrame vdpDebugFrame;		// VDP Debugger Frame

	private Setup setup;						// Pointer to General Parameters

	private String directory = "."; 			// Directory for Roms
	private boolean cart_loaded;				// Has a Cartridge Been Loaded

	private Image gameGear, splash;			// Images
	private JRadioButtonMenuItem ntscItem;		// Global, as must be selected for Game Gear


	/**
	 *	JavaGearFrame Constructor.
     */

	public JavaGearFrame()
	{
		// Load Images
		URL imageURL = JavaGear.class.getResource("images/gamegear.jpg");
		gameGear = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGear.class.getResource("images/icon.gif");
		Image icon = Toolkit.getDefaultToolkit().getImage(imageURL);
		imageURL = JavaGear.class.getResource("images/splash.jpg");
		splash = Toolkit.getDefaultToolkit().getImage(imageURL);

		// Wait for images to load
		MediaTracker tracker = new MediaTracker(this);
		tracker.addImage(gameGear, 0);
		tracker.addImage(icon, 1);
		tracker.addImage(splash, 2);
		try {tracker.waitForAll();}
		catch(InterruptedException e){}

		// Create pointer to general options
		setup = new Setup();
		
		// Initialise Objects
		controllers 		= new Controllers(setup);
		InterruptLine irq 	= new InterruptLine();
		mem  				= new Memory(setup);
		vdp 				= new Vdp(setup, irq);
		screenpanel 		= new Screen(setup, vdp);
		sn76496 			= new SN76496(111860.78125, 11025); // Clock Rate, Sample Rate
		port 				= new Ports(vdp, controllers, sn76496);
		z80cpu 				= new Z80(mem, port, irq);
		debugger			= new Debug(z80cpu, vdp, screenpanel);
		emulateLoop 		= new EmulateLoop(z80cpu, mem, vdp, sn76496, screenpanel);

		// Setup Frame
		setTitle("JavaGear");
		setIconImage(icon);

		// Ensure menus display over the emulation pane
		JPopupMenu.setDefaultLightWeightPopupEnabled(false);

		// Create Menu Bar
		JMenuBar menuBar = new JMenuBar();
		menuBar.add(makeMenu("File",
			new Object[]
			{
				"Open",
				"Close",
				null,
				"Exit"
			}, this));

		ButtonGroup regionGroup = new ButtonGroup();
		JRadioButtonMenuItem usItem = new JRadioButtonMenuItem("US/Europe");
		JRadioButtonMenuItem japItem = new JRadioButtonMenuItem("Japan");
		usItem.setSelected(true);
		regionGroup.add(usItem); regionGroup.add(japItem);

		ButtonGroup tvGroup = new ButtonGroup();
		ntscItem = new JRadioButtonMenuItem("NTSC");
		JRadioButtonMenuItem palItem = new JRadioButtonMenuItem("PAL");
		ntscItem.setSelected(true);
		tvGroup.add(ntscItem); tvGroup.add(palItem);

		ButtonGroup fskipGroup = new ButtonGroup();
		JRadioButtonMenuItem fs0 = new JRadioButtonMenuItem("1:1");
		JRadioButtonMenuItem fs1 = new JRadioButtonMenuItem("1:2");
		JRadioButtonMenuItem fs2 = new JRadioButtonMenuItem("1:3");
		fs0.setSelected(true);
		fskipGroup.add(fs0); fskipGroup.add(fs1); fskipGroup.add(fs2);

		ButtonGroup zoomGroup = new ButtonGroup();
		JRadioButtonMenuItem x1 = new JRadioButtonMenuItem("x 1");
		JRadioButtonMenuItem x2 = new JRadioButtonMenuItem("x 2");
		JRadioButtonMenuItem x3 = new JRadioButtonMenuItem("x 3");
		x1.setSelected(true);
		zoomGroup.add(x1); zoomGroup.add(x2); zoomGroup.add(x3);

		menuBar.add(makeMenu("System",
			new Object[]
			{
				"Controls",
				null,
				"Soft Reset",
				"Hard Reset",
				null,
				makeMenu("Region",
					new Object[]
					{
						usItem,
						japItem
					}, this),
				makeMenu("TV Type",
					new Object[]
					{
						ntscItem,
						palItem
					}, this)
			}, this));

		menuBar.add(makeMenu("Video",
			new Object[]
			{
				new JCheckBoxMenuItem("Show FPS", false),
				makeMenu("Frameskip",
					new Object[]
					{
						fs0,
						fs1,
						fs2
					}, this),
				makeMenu("Screen Size",
					new Object[]
					{
						x1,
						x2,
						x3,
					}, this),
					null,
				makeMenu("Layers",
					new Object[]
					{
						new JCheckBoxMenuItem("Background", true),
						new JCheckBoxMenuItem("Sprites", true),
					}, this),
			}, this));

		menuBar.add(makeMenu("Sound",
			new Object[]
			{
				new JCheckBoxMenuItem("Enable Sound", true),
				new JCheckBoxMenuItem("Record Sound", false),
				makeMenu("Channels",
					new Object[]
					{
						new JCheckBoxMenuItem("Tone Generator 1", true),
						new JCheckBoxMenuItem("Tone Generator 2", true),
						new JCheckBoxMenuItem("Tone Generator 3", true),
						new JCheckBoxMenuItem("Noise Generator " , true),
					}, this),

			}, this));

		menuBar.add(makeMenu("Debug",
			new Object[]
			{
				"Z80",
				"VDP",
			}, this));
		setJMenuBar(menuBar);

		// Set Window Size
		setup.setScreenSize(1);
		screenpanel.setSize(new Dimension(setup.h_end_scaled, setup.v_end_scaled));

		// Get Current Screen Size
		Toolkit tk = Toolkit.getDefaultToolkit();
		Dimension d = tk.getScreenSize();
		// Set Window in middle
		setLocation(d.width / 4, d.height / 4);

		// Add Window Listener
		addWindowListener(new WindowAdapter()
			{
				public void windowClosing(WindowEvent e) {terminate();}
		       });


		// Setup Emulation Defaults
		port.setEurope(true); 			// Default to a domestic system
		setup.setSMS(); 				// Default to SMS mode

		// Add Keyboard Support
		addKeyListener(controllers);

		// Add Display Window
		getContentPane().add(screenpanel);

		pack();
		setResizable(false);
		requestFocus();
	}


	/**
	 *	Paint Splash Screen and Game Gear Image
	 *
	 *  @param g	The graphics context to use for painting
     */

	public void paint(Graphics g)
	{
		super.paint(g);

		// Paint Game Gear Graphic
		if (cart_loaded)
		{
			if (setup.is_gg)
			{
				Dimension d = screenpanel.getSize();
				screenpanel.getGraphics().drawImage(gameGear, 0, 0, d.width, d.height, this);
				screenpanel.forceUpdate();
			}
			if (debugger.isVisible()) screenpanel.refresh();//fix for gg - chris
		}

		// Paint Splash Screen
		else
		{
			Dimension d = screenpanel.getSize();
			screenpanel.getGraphics().drawImage(splash, 0, 0, d.width, d.height, this);
		}
	}


	/**
	 *	Open Cartridge and Start Emulation
	 *
	 *  @param URL	URL of cartridge image
     */

	public void openCart(URL filename)
	{
		mem.reset();

		// Abort if errors reading file into memory
		if (mem.readCart(filename))
		{
			cart_loaded = true;
			emulateLoop.stopThread();

			// Set NTSC Timing if GameGear
			if (setup.is_gg)
			{
				emulateLoop.setNTSC();
				ntscItem.setSelected(true);
			}
			else
			{
				if (ntscItem.isSelected())
					emulateLoop.setNTSC();
				else
					emulateLoop.setPAL();
			}

			if (!debugger.isVisible()) emulateLoop.startThread();
			else emulateLoop.reset();
		}
		else
			cart_loaded = false;
	}


	/**
	 *	Invoke Dialog to select cartridge image.
	 *
	 *  @return		URL of chosen cartridge
     */

	public URL openDialog()
	{
		URL filename = null;

		// Create Frame for File Chooser
		JFrame openFrame = new JFrame("Open");

		// Create new File Chooser
		JFileChooser chooser = new JFileChooser();

		// Set Directory
		chooser.setCurrentDirectory(new File(directory));

		// Set FileFilters
		chooser.addChoosableFileFilter(new GGFilter());
		chooser.addChoosableFileFilter(new SMSFilter());
		chooser.setFileFilter(new RomFilter());
		chooser.setAcceptAllFileFilterUsed(false); // turn off *.* filter

		// This call does not return until a file is selected
		int result = chooser.showOpenDialog(openFrame);

		// Get the Filename
		File selectedFile = chooser.getSelectedFile();

		if (selectedFile != null)
		{
			//System.out.println("Chosen file is " + selectedFile.getPath());
			directory = selectedFile.getAbsolutePath();
			try
			{
				filename = selectedFile.toURL();
			}
			catch (MalformedURLException e){}

			return filename;
		}

		else return null;
	}


	/**
	 *	Invoked when a menu item is selected
     */

	public void actionPerformed(ActionEvent evt)
	{
		if (evt.getSource() instanceof JMenuItem)
		{
			String arg = evt.getActionCommand();

			// FILE MENU
			if (arg.equals("Open"))
			{
				if (!debugger.isVisible()) emulateLoop.suspendThread();

				// Get URL of file
				URL filename = openDialog();

				// Pass URL to openCart Method
				if  (filename != null)
					openCart(filename);
				else if (cart_loaded)
				{
					screenpanel.disposeGraphics();
					emulateLoop.resumeThread();
				}
			}

			else if (arg.equals("Close"))
			{
				emulateLoop.stopThread();
				mem.reset();
				cart_loaded = false;
			}

			else if (arg.equals("Exit"))
				terminate();

			// SOUND MENU
			else if (arg.equals("Record Sound"))
				sn76496.setRecord();
			else if (arg.equals("Enable Sound"))
				sn76496.setEnabled();
			else if (arg.equals("Tone Generator 1"))
				sn76496.setChannelEnabled(0);
			else if (arg.equals("Tone Generator 2"))
				sn76496.setChannelEnabled(1);
			else if (arg.equals("Tone Generator 3"))
				sn76496.setChannelEnabled(2);
			else if (arg.equals("Noise Generator "))
				sn76496.setChannelEnabled(3);

			// SYSTEM MENU
			else if (arg.equals("Controls"))
			{
				if (keyFrame == null) keyFrame = new KeyFrame(controllers);
				keyFrame.show();
			}
			else if (arg.equals("Hard Reset"))
			{
				if (cart_loaded)
				{
					if (!debugger.isVisible()) emulateLoop.stopThread();
					emulateLoop.reset();
					if (!debugger.isVisible()) emulateLoop.startThread();
				}
			}
			else if (arg.equals("Soft Reset"))
			{
				if (setup.is_sms)
					controllers.setReset();
			}

			else if (arg.equals("Show FPS"))
			{
				if (!show_fps)
				{
					show_fps = true;
					screenpanel.paintFPS(true);
				}
				else
				{
					show_fps = false;
					screenpanel.paintFPS(false);
				}
			}

			// VIDEO MENU
			else if (arg.equals("1:1"))
				Throttle.setFrameSkip(0);
			else if (arg.equals("1:2"))
				Throttle.setFrameSkip(1);
			else if (arg.equals("1:3"))
				Throttle.setFrameSkip(2);
			else if (arg.equals("Background"))
				vdp.setBackgroundLayer();
			else if (arg.equals("Sprites"))
				vdp.setSpriteLayer();

			else if (arg.equals("x 1"))
			{
				emulateLoop.suspendThread();
				setup.setScreenSize(1);
				screenpanel.disposeGraphics();
				screenpanel.setSize(new Dimension(setup.h_end_scaled, setup.v_end_scaled));
				pack();
				if (cart_loaded)
					emulateLoop.resumeThread();
			}
			else if (arg.equals("x 2"))
			{
				emulateLoop.suspendThread();
				setup.setScreenSize(2);
				screenpanel.disposeGraphics();
				screenpanel.setSize(new Dimension(setup.h_end_scaled, setup.v_end_scaled));
				pack();
				if (cart_loaded) emulateLoop.resumeThread();
			}
			else if (arg.equals("x 3"))
			{
				emulateLoop.suspendThread();
				setup.setScreenSize(3);
				screenpanel.disposeGraphics();
				screenpanel.setSize(new Dimension(setup.h_end_scaled, setup.v_end_scaled));
				pack();
				if (cart_loaded) emulateLoop.resumeThread();
			}
			else if (arg.equals("US/Europe"))
				port.setEurope(true);
			else if (arg.equals("Japan"))
				port.setEurope(false);
			else if (arg.equals("NTSC"))
			{
				emulateLoop.suspendThread();
				emulateLoop.setNTSC();
				if (cart_loaded) emulateLoop.resumeThread();
			}
			else if (arg.equals("PAL"))
			{
				if (setup.is_sms)
				{
					emulateLoop.suspendThread();
					emulateLoop.setPAL();
					if (cart_loaded) emulateLoop.resumeThread();
				}
				else
					ntscItem.setSelected(true);
			}

			// DEBUG MENU
			else if (arg.equals("Z80"))
			{
				if (cart_loaded)
					emulateLoop.stopThread();

				if (debugger == null)
					debugger = new Debug(z80cpu, vdp, screenpanel);

				debugger.show();
			}
			else if (arg.equals("VDP"))
			{
				if (vdpDebugFrame == null)
					vdpDebugFrame = new VdpDebugFrame(vdp);

				vdpDebugFrame.show();
			}
		}
	}


	/**
	 *	Terminate JavaGear
     */

	public void terminate()
	{
		System.out.print("Closing      ... ");
		sn76496.stopRecording();		// Close sound recording

		System.out.println("Done");
		System.gc();
		System.exit(0);
	}

	/**
		Building Menu Method.
		From: Core Java Volume 1 (1999 Sun Microsystems) P 489
	*/

	private static JMenu makeMenu(Object parent, Object[] items, Object target)
	{
		JMenu m = null;
		if (parent instanceof JMenu)
			m = (JMenu)parent;
		else if (parent instanceof String)
			m = new JMenu((String)parent);
		else
			return null;

		for (int i=0; i < items.length; i++)
		{
			if (items[i] == null)
				m.addSeparator();
			else
				m.add(makeMenuItem(items[i], target));
		}
		return m;
	}

	/**
		Building Menu Method.
		From: Core Java Volume 1 (1999 Sun Microsystems) P 489
	*/

	private static JMenuItem makeMenuItem(Object item, Object target)
	{
		JMenuItem r = null;
		if (item instanceof String)
			r = new JMenuItem((String)item);
		else if (item instanceof JMenuItem)
			r = (JMenuItem)item;
		else return null;

		if (target instanceof ActionListener)
			r.addActionListener((ActionListener)target);
		return r;
	}
}