/**
 * KeyConfig.java
 *
 * Create the Panel to Define Controller Configuration
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.awt.*;
import java.awt.event.*;	// Action Listeners
import java.awt.image.*;
import javax.swing.*;
import java.net.*; 		// URL Handling
import java.awt.geom.*; 	// Shapes


public final class KeyConfig extends JPanel
	implements MouseMotionListener, MouseListener, KeyListener
{
	private Graphics2D g2;			// Graphics Context
	private Image gameGear;		// Game Gear Image
	private Font f;
	private FontMetrics fm;
	private int x = 0, y = 0;		// Current Position of Mouse

	// Co-ordinates of Buttons for transparency
	private RoundRectangle2D button1 = new RoundRectangle2D.Double(295, 78, 20, 20, 30, 30);
	private RoundRectangle2D button2 = new RoundRectangle2D.Double(320, 58, 20, 20, 30, 30);
	private RoundRectangle2D buttonS = new RoundRectangle2D.Double(302, 32, 11, 18, 30, 30);
	private RoundRectangle2D buttonL = new RoundRectangle2D.Double(27, 72, 10, 13, 0, 0);
	private RoundRectangle2D buttonR = new RoundRectangle2D.Double(56, 72, 10, 14, 0, 0);
	private RoundRectangle2D buttonU = new RoundRectangle2D.Double(40, 60, 12, 12, 0, 0);
	private RoundRectangle2D buttonD = new RoundRectangle2D.Double(41, 87, 11, 12, 0, 0);
	private RoundRectangle2D buttonCurrent = null;

	// Current Key Configuration
	private int up = KeyEvent.VK_UP;
	private int down = KeyEvent.VK_DOWN;
	private int left = KeyEvent.VK_LEFT;
	private int right = KeyEvent.VK_RIGHT;
	private int fire1 = KeyEvent.VK_Z;
	private int fire2 = KeyEvent.VK_X;
	private int start = KeyEvent.VK_ENTER;

	// Is Mouse Clicked Inside a Button?
	private boolean mouse_pressed = false;

	public KeyConfig()
	{
		f = new Font("SansSerif", Font.PLAIN, 6);

		// Load Game Gear Image
		URL imageURL = KeyFrame.class.getResource("images/keyconfig.jpg");
		gameGear = Toolkit.getDefaultToolkit().getImage(imageURL);
		MediaTracker tracker = new MediaTracker(this);
		tracker.addImage(gameGear, 0);
		try {tracker.waitForID(0);}
		catch(InterruptedException e){}

		addKeyListener(this);
		addMouseMotionListener(this);
		addMouseListener(this);
	}

	public void setKeys(int [] newKeys)
	{
		up    = newKeys[0];
		down  = newKeys[1];
		left  = newKeys[2];
		right = newKeys[3];
		fire1 = newKeys[4];
		fire2 = newKeys[5];
		start = newKeys[6];
	}

	public int[] getKeys()
	{
		int[] keys = {up, down, left, right, fire1, fire2, start};
		return keys;
	}

	public void mouseMoved(MouseEvent evt)
	{
		x = evt.getX();
		y = evt.getY();
		
		if (buttonCurrent != null && !overButton(buttonCurrent))
			refresh();

		// Check to see if mouse is over one of the following areas
		if (!mouse_pressed)
		{
			if (overButton(button1) || 
				overButton(button2) ||
				overButton(buttonS) ||
				overButton(buttonL) ||
				overButton(buttonR) ||
				overButton(buttonU) ||
				overButton(buttonD)
				)
				refresh();
		}

		if (buttonCurrent != null)
			setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
		else
			setCursor(Cursor.getDefaultCursor());
	}

	public void mousePressed(MouseEvent evt)
	{
		if (buttonCurrent != null)
			mouse_pressed = true;
		else
			mouse_pressed = false;
		requestFocus();
		refresh();
	}

	public void mouseDragged(MouseEvent evt){} // for mouse motionlistener
	public void mouseClicked(MouseEvent evt){} // for mouse listener
	public void mouseReleased(MouseEvent evt){} // for mouse listener
	public void mouseEntered(MouseEvent evt){} // for mouse listener
	public void mouseExited(MouseEvent evt){} // for mouse listener

	public void keyTyped(KeyEvent evt){}
	public void keyReleased(KeyEvent evt){}
	public void keyPressed(KeyEvent evt)
	{
		// If mouse pressed, re-define the key
		if (mouse_pressed)
		{
			if (buttonCurrent.equals(button1))
				fire1 = evt.getKeyCode();
			else if (buttonCurrent.equals(button2))
				fire2 = evt.getKeyCode();
			else if (buttonCurrent.equals(buttonS))
				start = evt.getKeyCode();
			else if (buttonCurrent.equals(buttonU))
				up = evt.getKeyCode();
			else if (buttonCurrent.equals(buttonD))
				down = evt.getKeyCode();
			else if (buttonCurrent.equals(buttonL))
				left = evt.getKeyCode();
			else if (buttonCurrent.equals(buttonR))
				right = evt.getKeyCode();

			mouse_pressed = false;
		}
		refresh();
	}

	// Need the following to read from the keyboard
	//public boolean isFocusTraversable(){return true;}

	public boolean overButton(RoundRectangle2D button)
	{
		if ( (x >= button.getX()) && (x <= button.getX()+button.getWidth()) &&
			 (y >= button.getY()) && (y <= button.getY()+button.getHeight()) )
		{
			buttonCurrent = button;
			return true;
		}
		else
		{
			buttonCurrent = null;
			return false;
		}
	}

	public void refresh()
	{
		if (g2 == null) g2 = (Graphics2D) getGraphics();
		g2.drawImage(gameGear,0,0,this);

		if (buttonCurrent != null )
		{
			int rule = AlphaComposite.SRC_OVER;
			float alpha = (float) 0.5;	// Set transparency value
			g2.setComposite(AlphaComposite.getInstance(rule, alpha));
			g2.setPaint(Color.red);

			g2.fill(buttonCurrent);

			alpha = 1;
			g2.setComposite(AlphaComposite.getInstance(rule, alpha));
		}

			g2.setPaint(Color.white);
			fm = g2.getFontMetrics(f);
			int height = fm.getHeight()+2; // Distance between lines of text
			int textX = 133;
			int textY = 43;
			int offset = 50;

			g2.drawString("Up", textX, textY);
			g2.drawString("Down", textX, textY+=height);
			g2.drawString("Left", textX, textY+=height);
			g2.drawString("Right", textX, textY+=height);
			g2.drawString("Fire 1", textX, textY+=height);
			g2.drawString("Fire 2", textX, textY+=height);
			g2.drawString("Start", textX, textY+=height);

			textY = 43;	 // Reset Y
			textX += 45; // Offset X

			if (buttonCurrent != null)
				if (buttonCurrent.equals(buttonU))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);


			if ((mouse_pressed) && (buttonCurrent.equals(buttonU)))
				g2.drawString("Press Key", textX, textY);
			else
				g2.drawString(KeyEvent.getKeyText(up), textX, textY);

			if (buttonCurrent != null)
				if (buttonCurrent.equals(buttonD))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);

			if ((mouse_pressed) && (buttonCurrent.equals(buttonD)))
				g2.drawString("Press Key", textX, textY+=height);
			else
				g2.drawString(KeyEvent.getKeyText(down), textX, textY+=height);

			if (buttonCurrent != null)
				if (buttonCurrent.equals(buttonL))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);

			if ((mouse_pressed) && (buttonCurrent.equals(buttonL)))
				g2.drawString("Press Key", textX, textY+=height);
			else
				g2.drawString(KeyEvent.getKeyText(left), textX, textY+=height);

			if (buttonCurrent != null)
				if (buttonCurrent.equals(buttonR))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);

			if ((mouse_pressed) && (buttonCurrent.equals(buttonR)))
				g2.drawString("Press Key", textX, textY+=height);
			else
				g2.drawString(KeyEvent.getKeyText(right), textX, textY+=height);

			if (buttonCurrent != null)
				if (buttonCurrent.equals(button1))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);

			if ((mouse_pressed) && (buttonCurrent.equals(button1)))
				g2.drawString("Press Key", textX, textY+=height);
			else
				g2.drawString(KeyEvent.getKeyText(fire1), textX, textY+=height);

			if (buttonCurrent != null)
				if (buttonCurrent.equals(button2))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);

			if ((mouse_pressed) && (buttonCurrent.equals(button2)))
				g2.drawString("Press Key", textX, textY+=height);
			else
				g2.drawString(KeyEvent.getKeyText(fire2), textX, textY+=height);

			if (buttonCurrent != null)
				if (buttonCurrent.equals(buttonS))
					g2.setPaint(Color.red);
				else
					g2.setPaint(Color.white);

			if ((mouse_pressed) && (buttonCurrent.equals(buttonS)))
				g2.drawString("Press Key", textX, textY+=height);
			else
				g2.drawString(KeyEvent.getKeyText(start), textX, textY+=height);
	}

	public void paintComponent(Graphics g){}
	public void update(Graphics g) {}
	public void repaint(long tm, int x, int y, int width, int height){}
	public void paintAll(Graphics g) {}

	public Dimension getPreferredSize()
	{
		return new Dimension(gameGear.getWidth(this),gameGear.getHeight(this));
	}

	public Dimension getMinimumSize()
	{
		return getPreferredSize();
	}
}

