/**
 * Memory.java
 *
 * Emulates SMS/GG Memory and Paging Hardware.
 * Also Handles Reading ROMs and configuring the memory appropriately.
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.io.*;								// I/O Operations
import java.util.zip.*;						// ZIP file Support
import java.net.*;								// URL Support

public final class Memory
{
	private byte [][] frames;					// Array of Memory Frames
	private byte [][] pages;					// Array of Memory Pages
	private byte [][] cart_ram_pages;			// Array of Cartridge RAM Pages
	private boolean frame_two_rom;			// Stores whether frame 2 contains cartridge ROM or cartridge RAM
	private int cart_ram_page;				// Stores which Page of cartridge RAM to use
	private int number_of_pages;				// Stores number of pages

	private Setup setup;						// Pointer to General Parameters

	/**
	 *	Memory Constructor.
	 *
	 *  @param set	Pointer to General Parameters
     */

	public Memory(Setup set)
	{
		this.setup = set;
		// Create 4 x 16K Frames
		frames = new byte[4][0x4000];
		// Create 2 x 16K RAM Cartridge Pages
		cart_ram_pages = new byte[2][0x4000];
		// Create 4 x 16K Pages
		pages = new byte[4][0x4000];
		// Default number of pages
		number_of_pages = 2;
		// Reset to Default Values
		reset();
	}

	/**
	 *	Reset Memory to Default Values.
     */

	public void reset()
	{
		cart_ram_page = 0;
		frame_two_rom = true;

		// Default Mapping
		if (pages != null)
		{
			frames[0] = pages[0];
			frames[1] = pages[1];
		}
	}

	/**
	 *	Write to a memory location.
	 *
	 *  @param address		Memory address
	 *	@param value		Value to write
     */

	public void write(int address, int value)
	{
		// Do not write over ROM (GG Bartman tries this)
		if (address < 0x8000) return;		
		else if (address < 0xC000)	// ROM Page 2
		{
			if (!frame_two_rom)
				frames[2][address-0x8000]= (byte) value;
		}
		else if (address < 0xE000)	// RAM
			frames[3][address-0xA000]= (byte) value;
		else if (address < 0xFFFC)	// Mirror RAM
			frames[3][address-0xC000]= (byte) value;
		else if (address < 0x10000) // Paging registers!
			page(address, value);
	}

	/**
	 *	Read from a memory location.
	 *
	 *  @param address		Memory location
	 *
	 *	@return				Value from memory location
     */

	public int read(int address)
	{
		// Elegant :
		// return unsigned(readSigned(address));
		
		// Dirty Fast Solution (Copy Code Twice!)
		if (address < 0x400)		// First 1K
			return pages[0][address]&0xFF;
		else if (address < 0x4000)	// ROM Page 0
			return frames[0][address]&0xFF;
		else if (address < 0x8000)	// ROM Page 1
			return frames[1][address-0x4000]&0xFF;
		else if (address < 0xC000)	// ROM Page 2
			return frames[2][address-0x8000]&0xFF;
		else if (address < 0xE000)	// RAM
			return frames[3][address-0xA000]&0xFF;
		else if (address < 0x10000)// Mirror RAM
			return frames[3][address-0xC000]&0xFF;
		else
			return 0;		
	}

	/**
	 *	Read a signed value from a memory location.
	 *
	 *  @param address		Memory address
	 *
	 *	@return				Value from memory location
     */

	public int readSigned(int address)
	{
		if (address < 0x400)		// First 1K
			return pages[0][address];
		else if (address < 0x4000)	// ROM Page 0
			return frames[0][address];
		else if (address < 0x8000)	// ROM Page 1
			return frames[1][address-0x4000];
		else if (address < 0xC000)	// ROM Page 2
			return frames[2][address-0x8000];
		else if (address < 0xE000)	// RAM
			return frames[3][address-0xA000];
		else if (address < 0x10000)// Mirror RAM
			return frames[3][address-0xC000];
		else
			return 0;
	}

	/**
	 *	Read a word (two bytes) from a memory location.
	 *
	 *  @param address		Memory address
	 *
	 *	@return				Value from memory location
     */

	public int readWord(int address)
	{
		return ((read(address+1)<<8) + read(address));
	}

	/**
	 *	Unsign an Integer
	 *
	 *	@param b			Value to Unsign
	 *
	 *	@return				Unsigned Value
     */

	private int unsigned(int b)
	{
		return b & 0xFF;
	}

	/**
	 *	Write to a paging register
	 *
	 *	@param location		Memory location
	 *	@param value		Value to write
     */

	private void page(int address, int value)
	{
		switch(address)
		{
			// RAM/ROM select register
			case 0xFFFC:
				// Map RAM
				if ((value & 0x08) == 0x08)
				{
					frame_two_rom = false;
					if ((value & 0x04) == 0x04)
					{
						frames[2] = cart_ram_pages[1]; // Use second page
						cart_ram_page = 1;
					}
					else
					{
						frames[2] = cart_ram_pages[0]; // Use first page
						cart_ram_page = 0;
					}
				}
				// Map ROM
				else
					frame_two_rom = true;
				break;
			// Page 0 ROM Bank
			case 0xFFFD:
				frames[0] = pages[value%number_of_pages];
				break;
			// Page 1 ROM Bank
			case 0xFFFE:
				frames[1] = pages[value%number_of_pages];
				break;
			// Page 2 ROM Bank
			case 0xFFFF:
				if (frame_two_rom)
					frames[2] = pages[value%number_of_pages]; // Map ROM
				else
					frames[2] = pages[cart_ram_page];		  // Map RAM
				break;
		}

		frames[3][address-0xC000]= (byte) value;
	}

	/**
	 *	Read a cartridge into memory
	 *
	 *	@param url			URL of cartridge image
	 *
	 *	@return				True if cartridge read successfully
     */

	public boolean readCart(URL url)
	{
		int size = 0;				// File size
		boolean fail = false;		// Problems loading ROM

		URLConnection conn;
		BufferedInputStream bis;	// Buffered Input Stream
		ZipInputStream zip;			// Zip Input Stream

		try
		{
			//System.out.print("Opening ... "+url.toString());
			conn = url.openConnection();
			size = conn.getContentLength();			// Get File Size
			conn.getInputStream().close();

			bis = new BufferedInputStream(url.openStream());
			zip = new ZipInputStream(url.openStream());
		}
		catch(Exception e)
		{
			System.out.println("Error - Problem opening URL : ");
			System.out.println(e.getMessage());
			return false;
		}

		// Find out what sort of file we're dealing with
		String extension = getExtension(url.toString());

		// Determine file type and adjust emulation mode appopriately.

		// Set SMS Emulation On
		if (extension.equals("sms"))
			setup.setSMS();
		// Set GG Emulation On
		else if (extension.equals("gg"))
			setup.setGG();
		// ZIP File. (Need to do more work to determine emulation mode).
		else if (extension.equals("zip"))
		{
			try
			{
				// Has a SMS/GG File been found in zip file?
				boolean file_found = false;
				// Get next entry from the zip file
				ZipEntry entry = zip.getNextEntry();

				// cycle through zip until we find an SMS/GG file
				while ((entry != null) && (!file_found))
				{
					String extension2 = getExtension(entry.getName());
					// SMS/GG File Found
					if ((extension2.equals("sms")) || (extension2.equals("gg")))
					{
						// Set Emulation Mode
						if (extension2.equals("sms"))
							setup.setSMS();
						else
							setup.setGG();

						file_found = true;
						size = (int) entry.getSize();
						bis = new BufferedInputStream(zip);
						break;
					}
					// File is not SMS/GG File
					else
					{
						zip.closeEntry();
						entry = zip.getNextEntry();
					}
				}

				if (!file_found)
				{
					System.out.println("Error - No valid files found in zip file.");
					fail = true;
				}
			}
			catch(IOException e)
			{
				System.out.println("Error - Error in Zip File.");
				fail = true;
				return false;
			}
		} // end of zip file extraction

		if (fail) return false; // Failure in zip file somewhere...

		if (size < 0)
		{
			//System.out.println("Error - Could not obtain file size from server");
			return false;
		}
		//else if (size < 0x8000)
		//{
		//	System.out.println("Error - File is smaller than 32K! Aborting ");
		//	return false;
		//}

		try
		{
			// Strip 512 Byte File Headers
			if ((size % 1024) != 0)
			{
				bis.skip(512); // skip 512 bytes
				size -=512;
			}

			// Calculate number of pages from file size and create array appropriately
			number_of_pages = (int) (size/0x4000);
			pages = new byte[number_of_pages][0x4000];

			// Read file into pages array
			for (int x=0; x < size; x+=0x4000)
				// second value is offset, third is length
				bis.read(pages[x/0x4000], 0x0000, 0x4000);

			// Close all streams
			zip.close(); bis.close();

			// Default Mapping (Needed or Shinobi doesn't work)
			frames[0] = pages[0];
			frames[1] = pages[1];
		}

		catch(Exception e)
		{
			System.out.println("Error - File cannot be read or is corrupt");
			return false;
		}

		return true;
	}

	/**
	 *	Strip an extension from a file
	 *
	 *	@param filename		Filename to strip
	 *
	 *	@return				Extension of file
     */

	private String getExtension(String filename)
  	{
    	int i = filename.lastIndexOf('.');
    	if (i > 0 && i < filename.length() - 1)
      		return filename.substring(i+1).toLowerCase();
    	return "";
  	}
}