/**
 * NoiseGenerator.java
 *
 * Texas SN76496 Noise Generator Emulation.
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public final class NoiseGenerator
{
	/** Volume of channel */
	public int volume;

	/** Frequency of channel */
	public int frequency;

	/** First Byte of Frequency */
	public int first_byte;

	private double counter;					// Counter for square waves
	private boolean amplitude_flipflop;		// Positive/Negative output
	private double psg_cycles;				// PSG Cycles Per Sample
	private boolean enabled;					// Channel Enabled

	private int shift;						// Noise Source
	private final int RESET = 0x4000;			// Shift Register Reset Value SMS Specific (Bit 13)
	private final int WHITE_NOISE = 0x09;		// White Noise (values from Maxim)
	private final int SYNC_NOISE =  0x01; 	// Sync/Periodic Noise (values from Maxim)
	private int tapped_bits;				  	
	private int noise;						// Pointer to Noise

	private ToneGenerator chan2;				// Pointer to Channel 2 / Tone Generator 3
	private boolean use_chan2_freq;			// Use Channel 2's Frequency


	/**
	 *	NoiseGenerator Constructor.
	 *
	 *  @param clock_speed	Clock Speed (Hz)
	 *	@param sample_rate	Sample Rate (Hz)
	 *	@param t			Pointer to Tone Generator
     */

	public NoiseGenerator(double clock_speed, int sample_rate, ToneGenerator t)
	{
		this.chan2 = t;
		psg_cycles = (clock_speed / sample_rate)*2;
		reset();
	}


	/**
	 *	Reset Noise Generator to Default Values.
     */

	public void reset()
	{
		tapped_bits = WHITE_NOISE;
		volume = 0x0F;
		frequency = 1;
		first_byte = 0;
		counter = 0;
		noise = RESET;
		amplitude_flipflop = false;
		use_chan2_freq = false;
		enabled = true;
	}

	/**
	 *	Toggle Channel On/Off
     */

	public void setEnabled()
	{
		enabled = !enabled;
	}

	/**
	 *	Set Channel Frequency
	 *
	 *  @param	f		Frequency
     */

	public void setFrequency(int f)
	{
		// Bit 2 Selects White/Sync noise
		if ((f & 0x04) == 0x04)
			tapped_bits = WHITE_NOISE;
		else
			tapped_bits = SYNC_NOISE;

		// Reset shift register to preset
		shift = RESET;

		switch(f&0x03)
		{
			case 0x00:
				frequency = 0x10; // 16
				use_chan2_freq = false;
				break;
			case 0x01:
				frequency = 0x20; // 32
				use_chan2_freq = false;
				break;
			case 0x02:
				frequency = 0x40; // 64
				use_chan2_freq = false;
				break;
			// Channel 2 Frequency
			case 0x03:
				frequency = chan2.frequency;
				use_chan2_freq = true;
				break;
			default:
				break;
		}
	}


	/**
	 *	Return a single sample from noise Generator
	 *
	 *	@return		Sample (0 to +15)
     */

	public int getSample()
	{
		if (!enabled) return 0;

		// Decrement Counter
		counter -= psg_cycles;

		// Counter Expired
		if (counter < 0)
		{
			// Toggle Amplitude Flip Flop
			amplitude_flipflop = !amplitude_flipflop;;

			// Update Frequency from Channel 2
			if (use_chan2_freq) frequency = chan2.frequency;

			// Reload counter
			while (counter < 0)
			{
				if (frequency > 0)
					counter+=frequency;
				else
					counter=1; // so we don't get stuck in an endless loop
			}

			// We only want to do this once per cycle
			if (amplitude_flipflop)
				shift = (shift>>1) | (PARITY[(shift & tapped_bits) & 0x0F]<<15);
		}

		if ((shift&0x01) == 0x00)
			return 0; // Noise channel does not output negative values
		else
			return (15 - (volume&0x0f)); // ie 0 is 15, 1 is 14 etc
	}

	private final static int PARITY[] = { 1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1 };

}