/**
 * Ports.java
 *
 * A class to attach virtual devices to the Z80's ports.
 *
 * @author Copyright (C) 2002 Chris White
 * @version 6th March 2002
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public final class Ports
{
	private Vdp vdp;							// Pointer to VDP
	private Controllers controllers;			// Pointer to Controllers
	private SN76496 sn76496;					// Pointer to SN76496
	private int nat1;							// Nationalisation Bit 1
	private int nat2;							// Nationalisation Bit 2
	private boolean europe;						// European / Domestic System

	/**
	 *	Ports Constructor.
	 *
	 *  @param v	Vdp
     *  @param c	Controllers
     *  @param s	SN76496
     */

	public Ports(Vdp v, Controllers c, SN76496 s)
	{
		this.controllers = c;
		this.vdp = v;
		sn76496 = s;
		// Reset Nationalisation Bits
		nat1 = 0;
		nat2 = 0;
		// Default to a European System
		europe = true;
	}

	/**
	 *	Output to a Z80 port.
	 *
	 *  @param port		Port Number
     *  @param value	Value to Output
     */

	public void out(int port, int value)
	{
		switch(port)
		{
			// GG Serial Ports
			case 0x00:
			case 0x01:
			case 0x02:
			case 0x03:
			case 0x04:
			case 0x05:
				break;
			// VDP Data port
			case 0xBE:
				vdp.dataWrite(value);
				break;
			// VDP Control port (Mirrored at two locations)
			case 0xBD:
			case 0xBF:
				vdp.controlWrite(value);
				break;
			// Automatic Nationalisation
			case 0x3F:
				if ((value & 0x20) == 0x20) nat1 = 1;
				else nat1 = 0;

				if ((value & 0x80) == 0x80) nat2 = 1;
				else nat2 = 0;

				if (!europe)
				{
					nat1 = ~nat1;
					nat2 = ~nat2;
				}
				break;
			// SN76489 PSG Output (Mirrored)
			case 0x7E:
			case 0x7F:
				sn76496.write(value);
				break;
			default:
				break;
		}
	}

	/**
	 *	Read from a Z80 Port
	 *
	 *  @param port		Port Number
     *
     *	@return			Value from Port Number
     */

	public int in(int port)
	{
		int value = 0;

		switch(port)
		{
			// GameGear (Start Button and Nationalisation)
			case 0x00:
				value = controllers.getGameGearStart();
				if (europe) value |= 0x40;
				else value &= ~0x40;
				break;
			// GG Serial Communication Ports  -
			// Return 0 for now as "OutRun" gets stuck in a loop by returning 0xFF
			case 0x01:
			case 0x02:
			case 0x03:
			case 0x04:
			case 0x05:
				value = 0;
				break;
			// Controller 1 (mirrored)
			case 0xC0:
			case 0xDC:
				value = controllers.getController1();
				break;
			// Controller 2 (mirrored) and Nationalisation
			case 0xC1:
			case 0xDD:
				value = controllers.getController2();
				// Reads bits 6 and 7 to determine nationalisation
				if (nat1 == 1) value |= 0x40; else value &= ~0x40;
				if (nat2 == 1) value |= 0x80; else value &= ~0x80;
				break;
			// Vertical Port
			case 0x7E:
				value = vdp.getVCount();
				break;
			// Horizontal Port (NOT EMULATED)
			case 0x7F:
				//System.out.println("Read to horizontal port");
				break;
			// VDP Data port
			case 0xBE:
				value = vdp.dataRead();
				break;
			// VDP Control port
			case 0xBD:
			case 0xBF:
				value = vdp.controlRead();
				break;
			// Default Value is 0xFF
			default:
				value = 0xff;
				break;
		}

		return value;
	}

	/**
	 *	Set Console to European / Japanese Model
	 *
	 *  @param value	True is European, False is Japanese
     */

	public void setEurope(boolean value)
	{
		europe = value;
	}
}
