/**
 * Registers.java
 *
 * Zilog Z80 CPU BC, DE, HL, IX, IY Register Pair Emulation
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public final class Registers
{
	private int high;
	private int low;

	private int high2;
	private int low2;

	private Flag f;


	/**
	 *	Register Pair Constructor
	 *
	 *	@param	fl	Pointer to Flag Register
     */

	public Registers(Flag fl)
	{
		high  = 0;
		high2 = 0;
		low   = 0;
		low2  = 0;
		this.f = fl;
	}


	/**
	 *  Clear Register
	 */

	public void clear()
	{
		high  = 0;
		high2 = 0;
		low   = 0;
		low2  = 0;
	}

	/**
	 *	Get Value of Register Pair
	 *
	 *	@return				Value Stored in Register Pair
     */

	public int get()
	{
		return (high<<8)|low;
	}

	/**
	 *	Get High Byte of Pair
	 *
	 *	@return				Value Stored in Register
     */

	public int getH() {return high;}


	/**
	 *	Get Low Byte of Pair
	 *
	 *	@return				Value Stored in Register
     */

    public int getL()  {return low;}

	/**
	 *	Set Register Pair
	 *
	 *	@param value			Value to set Register (0 - 0xFFFF)
     */

	public void set(int value)
	{
		low   = value&0xff;
		high  = (value>>8)&0xff;
	}

	/**
	 *	Set High Byte of Register Pair
	 *
	 *	@param value		Value to set Register (0 - 0xFF)
     */

	public void setH(int value)
	{
		high = value&0xff;
	}

	/**
	 *	Set Low Byte of Register Pair
	 *
	 *	@param value		Value to set Register (0 - 0xFF)
     */

	public void setL(int value)
	{
		low = value&0xff;
	}

	/**
	 *	Increment Register Pair
     */

	public void inc()
	{
		low = (low + 1)&0xff;
		// 16 bit register so check high is zero before adding to low
		if (low == 0) high = (high + 1)&0xff;

	}

	/**
	 *	Increment Low Byte Of Pair
     */

	public void incL()  {low  = inc8(low);}

	/**
	 *	Increment High Byte Of Pair
     */

	public void incH() {high = inc8(high);}
	private int inc8(int value)
	{
		value = (value + 1)&0xff;

		if ((value & 0x0F) == 0x00)
			f.halfCarryOn();
		else
			f.halfCarryOff();

		if (value == 0x80) f.parityOn();
		else f.parityOff();

		f.negativeOff();

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) == 0) f.signOff();
		else f.signOn();

		return value;
	}

	/**
	 *	Decrement Register Pair
     */

	public void dec()
	{
		low = (low - 1)&0xff;
		// 16 bit register so if high is full then decrease low too
		if (low == 255) high = (high - 1)&0xff;
	}

	/**
	 *	Decrement Low Byte Of Pair
     */

	public void decL()  {low  = dec8(low);}

	/**
	 *	Decrement High Byte Of Pair
     */

	public void decH() {high = dec8(high);}
	private int dec8(int value)
	{
		value = (value - 1)&0xff;

		if ((value & 0x0F) == 0x0F)
			f.halfCarryOn();
		else
			f.halfCarryOff();

		if (value == 0x7F) f.parityOn();
		else f.parityOff();

		f.negativeOn();

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) == 0) f.signOff();
		else f.signOn();

		return value;
	}


	/**
	 * Exchange Register Banks
	 */

	public void ex()
	{
		int temp;

		temp = high; high = high2; high2 = temp;
		temp = low; low = low2; low2 = temp;
	}


	/**
	 *	ADD 16 BIT (Verified with ZEXALL)
	 *
	 *  @param value		Value to add
     */

	public void add(int value)
	{
		int reg = get();
		int result = reg + value;

		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x1000) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x10000) != 0) f.carryOn();
	    else f.carryOff();

		result &= 0xFFFF;

		set(result);

		f.negativeOff();

		result>>=8;			// Bits 3 and 5 affected by high byte addition
		f.setBit3(result);
		f.setBit5(result);
	}


	/**
	 *	ADC 16 BIT - Add with carry
	 *
	 *  @param value		Value to add
     */
	public void adc(int value)
	{
		int reg = get();
		int result = reg + value + (f.carry() ? 1 : 0);
		
		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x1000) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x10000) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) f.parityOn();
	    else f.parityOff();
	    
	    // truncate
	    result &= 0xFFFF;

		// zero	    
		if (result == 0) f.zeroOn();
		else f.zeroOff();

		// sign
		if ((result & 0x8000) == 0) f.signOff();
		else f.signOn();
	
		// negative
		f.negativeOff();
		
		set(result);

		result>>=8;			// Bits 3 and 5 affected by high byte addition
		f.setBit3(result);
		f.setBit5(result);
	}


	/**
	 *	SBC 16 BIT - Subtract with carry
	 *
	 *  @param value		Value to subtract
     */

	public void sbc(int value)
	{
		int reg = get();
		int result = reg - value - (f.carry() ? 1 : 0);
				
		int carrybits = reg ^ value ^ result;
		
		// half carry
		if ((carrybits & 0x1000) != 0 ) f.halfCarryOn();
	    else f.halfCarryOff();
	    
	    // carry
	    if ((carrybits & 0x10000) != 0) f.carryOn();
	    else f.carryOff();
	    
	    // overflow
	    if ((((carrybits << 1) ^ carrybits) & 0x10000) != 0) f.parityOn();
	    else f.parityOff();
	    
	    // truncate
	    result &= 0xFFFF;

		// zero	    
		if (result == 0) f.zeroOn();
		else f.zeroOff();

		// sign
		if ((result & 0x8000) == 0) f.signOff();
		else f.signOn();
	
		// negative
		f.negativeOn();
		
		set(result);
		
		result>>=8;			// Bits 3 and 5 affected by high byte addition
		f.setBit3(result);
		f.setBit5(result);
	}
}