/**
 * SN76496.java
 *
 * Texas SN76496 Emulation.
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import javax.sound.sampled.*; 						// Java Sound API
import java.io.*; 									// Recording Sound

public final class SN76496 //extends Thread
{
	private ToneGenerator chan0;					// Tone Generator 1
	private ToneGenerator chan1;					// Tone Generator 2
	private ToneGenerator chan2;					// Tone Generator 3
	private NoiseGenerator chan3;					// Noise Generator

	private ToneGenerator current_generator;		// Pointer to current Tone Generator

	private byte [] buffer; 						// Buffer for sound data
	private FileWriter fileWriter;					// For Recording Sound to Disk

	private AudioFormat audioFormat;
	private SourceDataLine line; 					// The line that transmits audio to speakers

	private double clock_speed;					// PSG Clock Speed
	private int sample_rate;						// Output Sample Rate
	private int samples_per_frame;				// Samples to Generate per video frame

	private boolean enabled;						// Sound Enabled
	private boolean recording;					// Record Sound to Disk
	


	/**
	 *	SN76496 Constructor.
	 *
	 *  @param c	Clock Speed (Hz)
	 *	@param s	Sample Rate (Hz)
     */

	public SN76496(double c, int s)
	{
		clock_speed = c;
		sample_rate = s;

		chan0 = new ToneGenerator(clock_speed, sample_rate);
		chan1 = new ToneGenerator(clock_speed, sample_rate);
		chan2 = new ToneGenerator(clock_speed, sample_rate);
		chan3 = new NoiseGenerator(clock_speed, sample_rate, chan2);

		// AudioFormat(float sample_rate(hz), int sampleSizeInBits, int channels, boolean signed, boolean bigEndian)
		audioFormat = new AudioFormat(sample_rate, 8, 1, true, false);
		DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);

		if (!AudioSystem.isLineSupported(info))
			System.out.print("Audio not supported...");
		else
		{
			try
			{
				line = (SourceDataLine) AudioSystem.getLine(info);
				line.open(audioFormat);
				line.start();
			}
			catch (LineUnavailableException e)
			{
				System.out.print("Audio Line Unavailable...");
			}
		}

		recording = false;

		// Default to 60 FPS
		setFPS(60);

		// Reset to Defaults
		reset();
		
		// Enabled by Default
		setEnabled();
		
		
	}


	/**
	 *	Reset SN76496 to Default Values.
     */

	public void reset()
	{
		current_generator = chan0;
		chan0.reset();
		chan1.reset();
		chan2.reset();
		chan3.reset();
	}


	/**
	 *	Is SN76496 Enabled?
	 *
	 *	@return		True if enabled.
     */

	public boolean isEnabled()
	{
		return enabled;
	}


	/**
	 *	Toggle SN76496 On/Off
	 */

	public void setEnabled()
	{
		if (enabled)
		{
			stopSound();
		}
		else
		{
			startSound();
		}
	}


	/**
	 *	Toggle Particular SN76496 Channel On/Off
	 *
	 *	@param channel	Channel to Toggle (0-3)
	 */

	public void setChannelEnabled(int channel)
	{
		switch (channel)
		{
			case 0:
				chan0.setEnabled();
				break;
			case 1:
				chan1.setEnabled();
				break;
			case 2:
				chan2.setEnabled();
				break;
			case 3:
				chan3.setEnabled();
				break;
			default:
				break;
		}
	}

	/**
	 *	Set current FPS Rate.
	 *  Required to generate the correct number of samples per frame.
	 *
	 *  @param	v	FPS Rate
     */

	public void setFPS(int v)
	{
		samples_per_frame = (sample_rate / v);
		buffer = new byte[samples_per_frame];
		
	}


	/**
	 *	Program the PSG. Connected this procedure to a Z80 Port.
	 *
	 *  @param	value	Value to write (0-0xFF)
     */

	public void write(int value)
	{
		if ((!enabled) && (!recording)) return;

		if ((value & 0x80) == 0x80)
		{
			switch((value>>4) & 0x07)
			{
				case 0x00: // 000 rr = 00 (Channel 0 Frequency)
					chan0.first_byte = value&0x0F;
					current_generator = chan0;
					break;
				case 0x01: // 001 rr = 00 (Channel 0 Volume)
					chan0.volume = value&0x0F;
					current_generator = chan0;
					break;
				case 0x02: // 010 rr = 01 (Channel 1 Frequency)
					chan1.first_byte = value&0x0F;
					current_generator = chan1;
					break;
				case 0x03: // 011 rr = 01 (Channel 1 Volume)
					chan1.volume = value&0x0F;
					current_generator = chan1;
					break;
				case 0x04: // 100 rr = 10 (Channel 2 Frequency)
					chan2.first_byte = value&0x0F;
					current_generator = chan2;
					break;
				case 0x05: // 101 rr = 10 (Channel 2 Volume)
					chan2.volume = value&0x0F;
					current_generator = chan2;
					break;
				case 0x06: // 110 rr = 11 (Noise Channel Frequency)
					chan3.setFrequency(value&0x07);
					current_generator = null;
					break;
				case 0x07: // 111 rr = 11 (Noise Channel Volume)
					chan3.volume = value&0x0F;
					current_generator = null;
					break;
				default:
					break;
			}
		}
		// Set significant bits of frequency
		else if (current_generator != null)
			current_generator.setFreqSigf(value);
		else
			chan3.setFrequency(value&0x07);
	}
	
	/**
	 *	Start Emulation
     */

	public synchronized void startSound()
	{
		reset();
		enabled = true;
		line.start();
	}

	/**
	 *	Stop Emulation
     */

	public synchronized void stopSound()
	{
		enabled = false;
		line.stop();
	}


	/**
	 *	Convert PSG Settings to Java Sound
     */

	public void output()
	{
		if (!enabled && !recording) return;
	
		// Loop for length of this video frame
		for (int i = 0; i < samples_per_frame; i++)
		{
			// Sum the channel outputs
			int join = 0;
			join +=chan0.getSample();
			join +=chan1.getSample();
			join +=chan2.getSample();
			join +=chan3.getSample();

			// Scale volume up
			join <<= 1;

			// Check boundaries
			if (join > 0x7F) join = 0x7F;
			else if (join < -0x80) join = -0x80;
			
			buffer[i] = (byte) join;
			
			if (recording)
			{
				try
				{
					fileWriter.write(join&0xff); // output 8 bit signed mono
				}
				catch(IOException e){}
			}			
		}
		
		// Write to Java Line
		if (enabled)
		{
			 try
			 {
				 // Output Stream write(byte[] b, int off, int len)
				 // Small buffer to avoid latency, but more intensive CPU usage				 
				 line.write(buffer,0,samples_per_frame);
	         }
	         catch (Exception e) {}
		}
	}
	



	/**
	 *	Place Data in Java Sound Buffer
	 *
	 *  @param	value	Value to write
     */

	/*private void lineWrite()
	{
		 //if (line == null) return;
		 //buffer[0] = (byte) (value);

		 try
		 {
			 // Output Stream write(byte[] b, int off, int len)
			 // Small buffer to avoid latency, but more intensive CPU usage
			 line.write(buffer,0,1);
         }
         catch (Exception e) {}
	}*/


	/**
	 *	Toggle sound recording to WAV file.
     */

	public void setRecord()
	{
		if (recording)
			stopRecording();
		else
			startRecording();
	}


	/**
	 *	Start sound recording to WAV file.
     */

	private void startRecording()
	{
		if (!recording)
		{
			try
			{
				fileWriter = new FileWriter("output.raw");
			}
			catch(IOException e)
			{
				System.out.println("Could not open file for recording");
			}
			recording = true;
		}
	}


	/**
	 *	Stop sound recording to WAV file.
     */

	public void stopRecording()
	{
		if (recording)
		{
			try
			{
				fileWriter.close();
				convertToWav();
			}
			catch(IOException e)
			{
				System.out.println("Failed whilst closing output.raw");
			}
			recording = false;
		}
	}

	/**
	 *	Convert RAW output to WAV file.
     */

	private void convertToWav()
	{
		File input  = new File("output.raw");
		File output = new File("output.wav");

		try
		{
			FileInputStream fileInputStream = new FileInputStream(input);

			AudioInputStream audioInputStream = new AudioInputStream(fileInputStream, audioFormat, input.length());
			AudioSystem.write(audioInputStream, AudioFileFormat.Type.WAVE, output);
			audioInputStream.close();
			input.delete(); // Delete old file
			System.out.println("OUTPUT.WAV recorded");
		}
		catch (Exception e)
		{
			System.out.println("Error writing WAV file");
		}
	}
}