/**
 * ScreenPanel.java
 *
 * Component to draw the emulated display.
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
    This file is part of JavaGear.

    JavaGear is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    JavaGear is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with JavaGear; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

import java.awt.*;

public final class Screen extends Component
{		
	private Graphics gfx;				// Pointer to Graphics Context
	private boolean show_fps;			// Display FPS Counter
	private String fps;				// FPS String
	private Font f;					// Font to use
	
	private FastImage fast_image;		// Pointer to Fast Image Consumer
	private boolean pixel_update[];	// Pixel has been updated?

	private Vdp vdp;					// Pointer to VDP
	private Setup setup;				// Pointer to General Parameters
	
	private boolean force_update;


	/**
	 *	ScreenPanel Constructor.
	 *
	 *  @param set	Pointer to General Parameters
	 *  @param v 	Pointer to VDP
     */

	public Screen(Setup set, Vdp v)
	{
		this.setup   = set;	
		this.vdp = v;
		
		fast_image = new FastImage(setup.SMS_WIDTH, 1);	
		pixel_update = new boolean[setup.SMS_HEIGHT];

		// Do not display FPS Counter by default
		fps   = "";
		show_fps = false;

		// Font for FPS Counter
		f = new Font("SansSerif", Font.BOLD, 12);
		
	}

	/**
	 *	Refresh the screen.
     */

	public void refresh()
	{
		// Create Graphics Context for Component
		if (gfx == null) gfx = getGraphics();

		// SMS : Draw and Scale Image
		if (setup.is_sms)
		{	
			// Flag Changes Since Last Screen Redraw
			vdp.flagChanges(pixel_update);
			
			// Draw each line and scale individually
			// This yielded a speed increase over drawing one single 256x192 image
			for (int y = setup.SMS_HEIGHT; y-- != 0;)
			{
				// Check line has been updated
				if (pixel_update[y] || force_update)
				{
					fast_image.update(vdp.display, (y<<8));
					fast_image.paint(gfx, 0, y * setup.scale, setup.h_end_scaled, setup.scale);
				}	
			}
		}

		// GG  : Draw and Scale Centered Image
		else if (setup.is_gg)
		{
			vdp.flagChanges(pixel_update);
			
			for (int y = setup.GG_Y_OFFSET; y < setup.v_end; y++)
			{
				if (pixel_update[y] || force_update)
				{
					fast_image.update(vdp.display, (y<<8));
					fast_image.paint(gfx,setup.h_start_scaled, y*setup.scale,
										 setup.h_end_scaled_gg, y*setup.scale+setup.scale,
										 setup.GG_X_OFFSET, 0,
									 	 setup.h_end, 1);
				}	
			}			
		}
			
		if (force_update)
			force_update = false;
						
		// Paint FPS Counter
		if (show_fps)
		{
			// PAINT RECTANGLE
			gfx.setColor(Color.blue);
			gfx.fillRect(0, 0, 86, 20);

			// PAINT FPS TEXT
			gfx.setFont(f);
			gfx.setColor(Color.white);
			//gfx.drawString(fps, 1, 15);
			gfx.drawString(getFPS(), 1, 15);
			
		}
	}

	/**
	 *	Paint FPS Counter
	 *
	 *  @param b	True to paint, False to not paint
     */

	public void paintFPS(boolean b)
	{
		show_fps = b;
	}


	/**
	 *	Get Frames Per Second
	 *
	 *  @return	Average Frames Rendered Per Second, with corresponding percentage.
     */

	private String getFPS()
	{
		return  ((int)Throttle.getAverageFPS()+" FPS  ("+Throttle.getPercentage()+"%)");
	}


	/**
	 *	Repaint Screen When Window Obscured
	 * 
	 *  @param g	The graphics context to use for painting
     */

	public void paint(Graphics g)
	{			
		if (setup.is_sms)
		{
			for (int y = setup.SMS_HEIGHT; y-- != 0;)
			{
				fast_image.update(vdp.display, (y<<8));
				fast_image.paint(g, 0, y * setup.scale, setup.h_end_scaled, setup.scale);	
			}
		}
		else if (setup.is_gg)
		{
			for (int y = setup.GG_Y_OFFSET; y < setup.v_end; y++)
			{
				fast_image.update(vdp.display, (y<<8));
				fast_image.paint(g,  setup.h_start_scaled, y*setup.scale,
									 setup.h_end_scaled_gg, y*setup.scale+setup.scale,
									 setup.GG_X_OFFSET, 0, 
								 	 setup.h_end, 1);
			}	
		}
	}
	
	/**
	 * Force Screen To Update regardless of whether pixel changes have been made 
	 */
	
	public void forceUpdate()
	{
		force_update = true;	
	}


	/**
	 *	Overridden For Optimisation. Does Nothing.
	 *  @param g	The graphics context to use for painting
     */

	public void paintAll(Graphics g) {}


	/**
	 *	Overridden For Optimisation. Does Nothing.
	 *  @param g	The graphics context to use for painting
     */

	public void update(Graphics g) {}


	/**
	 *	Overridden For Optimisation. Does Nothing.
	 */

	public void repaint(){}

	/**
	 *	Overridden For Optimisation. Does Nothing.
		@param	x - 	the x coordinate.
		@param	y - 	the y coordinate.
		@param	width - the width.
		@param	height -the height.
	*/

	public void repaint(int x, int y, int width, int height){}


	/**
	 *	Overridden For Optimisation. Does Nothing.
	 	@param  tm - 	maximum time in milliseconds before update
		@param	x - 	the x coordinate.
		@param	y - 	the y coordinate.
		@param	width - the width.
		@param	height -the height.
	*/

	public void repaint(long tm, int x, int y, int width, int height){}


	/**
	 *	Dispose of graphics context. Necessary to resize the screen.
	 */

	public void disposeGraphics()
	{
		gfx = null;
		force_update = true;
	}


	/**
	*  Gets the preferred size of this component.
	*
	*  @return		A dimension object indicating this component's preferred size.
	*/

    public Dimension getPreferredSize()
    {
		return getSize();
	}


	/**
	*  Gets the mininimum size of this component.
	*
	*  @return		A dimension object indicating this component's minimum size.
	*/

   	public Dimension getMinimumSize()
    {
		return getPreferredSize();
	}
}
