/**
 * ToneGenerator.java
 *
 * Texas SN76496 Tone Generator Emulation.
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

public final class ToneGenerator
{
	/** Volume of channel */
	public int volume;

	/** Frequency of channel */
	public int frequency;

	/** First Byte of Frequency */
	public int first_byte;

	private double counter;					// Counter for square waves
	private boolean amplitude_flipflop;		// Positive/Negative output
	private double psg_cycles;				// PSG Cycles Per Sample
	private boolean enabled;					// Channel Enabled


	/**
	 *	ToneGenerator Constructor.
	 *
	 *  @param clock_speed	Clock Speed (Hz)
	 *	@param sample_rate	Sample Rate (Hz)
     */

	public ToneGenerator(double clock_speed, int sample_rate)
	{
		psg_cycles = ((clock_speed / sample_rate)*2); // *2 as we cycle for half wave
		enabled = true;
		reset();
	}


	/**
	 *	Reset Tone Generator to Default Values.
     */

	public void reset()
	{
		amplitude_flipflop = true;
		volume = 0x0F;
		frequency = 1;
		first_byte = 0;
		counter = 0;
	}


	/**
	 *	Toggle Channel On/Off
     */

	public void setEnabled()
	{
		enabled = !enabled;
	}
	
	/**
	 * Set Significant bits of frequency
	 * 
	 * @param value	High Byte
	 *
	 */

	public void setFreqSigf(int value)
	{
		frequency = first_byte | ((value&0x3F)<<4);
	}

	/**
	 *	Return a single sample from Tone Generator
	 *
	 *	@return		Sample (-15 to +15)
     */

	public int getSample()
	{
		if (!enabled) return 0;

		// Decrement Counter
		counter -= psg_cycles;

		// Counter Expired
		if (counter < 0)
		{
			// Toggle Amplitude Flip Flop
			amplitude_flipflop = !amplitude_flipflop;

			// Reload counter
			while (counter < 0)
			{
				if (frequency > 0)
					counter+=frequency;
				else
					counter=1; // so we don't get stuck in an endless loop
			}
		}

		int output_volume = 15 - (volume&0x0F); // ie 0 is 15, 1 is 14 etc

		// Real SMS Only Outputs Values Over 4?
		if (frequency > 4)
		{
			if (amplitude_flipflop)
				return +output_volume;
			else
				return -output_volume;
		}
		else
			return 0;
	}
}