/**
 * Z80.java
 *
 * Zilog Z80 CPU Emulation
 *
 * @author Copyright (C) 2002-2003 Chris White
 * @version 18th January 2003
 * @see "JavaGear Final Project Report"
 */

 /*
   This file is part of JavaGear.

   JavaGear is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   JavaGear is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with JavaGear; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/**
   For generic Z80 use the following need to be implemented correctly:
   + Interrupt Modes 0 and 2
   + Certain undocumented opcodes 
   + Make stack pointer wrap properly
*/

public final class Z80
{
	private int pc = 0x0000;					// Program Counter
	private int sp = 0x0000;					// Stack Pointer

	private int im = 0;						// Interrupt Mode (0,1,2)
	private boolean iff1 = false;			// Interrupt Flip Flop 1
	private boolean iff2 = false;			// Interrupt Flip Flop 2

	private boolean halt    = false; 		// Halt Instruction
	private boolean EI_inst = false;			// Previous Instruction was EI

	private InterruptLine irq;					// Pointer to Z80 Interrupt Line
	private Memory mem;						// Pointer to System Memory
	private Ports port;						// Pointer to Z80 Ports

												// ----- REGISTERS -----
	private Flag f;							// Flag Register
	private Accumulator a;						// Accumulator
	private Registers bc;						// Loop Counter
	private Registers de;						// General Purpose
	private Registers hl;						// General Purpose
	private Registers ix;						// Index Page Address Register
	private Registers iy;						// Index Page Address Register
	private Refresh r;							// Memory Refresh Register
	private int i;							// Interrupt page address register

	private int daa[];						// Pre-calculated result for DAA instruction

	private int tstates;						// T States

												// ------ OPCODE TIMINGS -----

	private final short opstates[] = {

										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */ 	4,10, 7, 6, 4, 4, 7, 4, 4,11, 7, 6, 4, 4, 7, 4,
										/* 0x10 */  8,10, 7, 6, 4, 4, 7, 4,12,11, 7, 6, 4, 4, 7, 4,
										/* 0x20 */  7,10,16, 6, 4, 4, 7, 4, 7,11,16, 6, 4, 4, 7, 4,
										/* 0x30 */  7,10,13, 6,11,11,10, 4, 7,11,13, 6, 4, 4, 7, 4,
										/* 0x40 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x50 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x60 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x70 */  7, 7, 7, 7, 7, 7, 4, 7, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x80 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0x90 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0xA0 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0xB0 */  4, 4, 4, 4, 4, 4, 7, 4, 4, 4, 4, 4, 4, 4, 7, 4,
										/* 0xC0 */  5,10,10,10,10,11, 7,11, 5,10,10, 0,10,17, 7,11,
										/* 0xD0 */  5,10,10,11,10,11, 7,11, 5, 4,10,11,10, 0, 7,11,
										/* 0xE0 */  5,10,10,19,10,11, 7,11, 5, 4,10, 4,10, 0, 7,11,
										/* 0xF0 */  5,10,10, 4,10,11, 7,11, 5, 6,10, 4,10, 0, 7,11 };

	private final short opcbstates[] = {

										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */ 	8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x10 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x20 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x30 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x40 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x50 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x60 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x70 */  8, 8, 8, 8, 8, 8,12, 8, 8, 8, 8, 8, 8, 8,12, 8,
										/* 0x80 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0x90 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xA0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xB0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xC0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xD0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xE0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8,
										/* 0xF0 */  8, 8, 8, 8, 8, 8,15, 8, 8, 8, 8, 8, 8, 8,15, 8 };

	private final short opddstates[] = {

										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */  4, 4, 4, 4, 4, 4, 4, 4, 4,15, 4, 4, 4, 4, 4, 4,
										/* 0x10 */  4, 4, 4, 4, 4, 4, 4, 4, 4,15, 4, 4, 4, 4, 4, 4,
										/* 0x20 */  4,14,20,10, 8, 8,11, 4, 4,15,20,10, 8, 8,11, 4,
										/* 0x30 */  4, 4, 4, 4,23,23,19, 4, 4,15, 4, 4, 4, 4, 4, 4,
										/* 0x40 */  4, 4, 4, 4, 8, 8,19, 4, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0x50 */  4, 4, 4, 4, 8, 8,19, 4, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0x60 */  8, 8, 8, 8, 8, 8,19, 8, 8, 8, 8, 8, 8, 8,19, 8,
										/* 0x70 */ 19,19,19,19,19,19, 4,19, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0x80 */  4, 4, 4, 4, 8, 8,19, 4, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0x90 */  4, 4, 4, 4, 8, 8,19, 4, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0xA0 */  4, 4, 4, 4, 8, 8,19, 4, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0xB0 */  4, 4, 4, 4, 8, 8,19, 4, 4, 4, 4, 4, 8, 8,19, 4,
										/* 0xC0 */  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
										/* 0xD0 */  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
										/* 0xE0 */  4,14, 4,23, 4,15, 4, 4, 4, 8, 4, 4, 4, 4, 4, 4,
										/* 0xF0 */  4, 4, 4, 4, 4, 4, 4, 4, 4,10, 4, 4, 4, 4, 4, 4};

	private final short opindexcbstates[] = {

										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0x10 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0x20 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0x30 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0x40 */ 00,00,00,00,00,00,20,00,00,00,00,00,00,00,20,00,
										/* 0x50 */ 00,00,00,00,00,00,20,00,00,00,00,00,00,00,20,00,
										/* 0x60 */ 00,00,00,00,00,00,20,00,00,00,00,00,00,00,20,00,
										/* 0x70 */ 00,00,00,00,00,00,20,00,00,00,00,00,00,00,20,00,
										/* 0x80 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0x90 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0xA0 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0xB0 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0xC0 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0xD0 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0xE0 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00,
										/* 0xF0 */ 00,00,00,00,00,00,23,00,00,00,00,00,00,00,23,00 };

	private final short opedstates[] = {

										/*          0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F */
										/* 0x00 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x10 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x20 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x30 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x40 */ 12,12,15,20, 8,14, 8, 9,12,12,15,20, 8,14, 8, 9,
										/* 0x50 */ 12,12,15,20, 8,14, 8, 9,12,12,15,20, 8,14, 8, 9,
										/* 0x60 */ 12,12,15,20, 8,14, 8,18,12,12,15,20, 8,14, 8,18,
										/* 0x70 */  8,12,15,20, 8,14, 8, 8,12,12,15,20, 8,14, 8, 8,
										/* 0x80 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0x90 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xA0 */ 16,16,16,16, 8, 8, 8, 8,16,16,16,16, 8, 8, 8, 8,
										/* 0xB0 */ 16,16,16,16, 8, 8, 8, 8,16,16,16,16, 8, 8, 8, 8,
										/* 0xC0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xD0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xE0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
										/* 0xF0 */  8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8 };


	/**
	 *	Z80 Constructor.
	 *
	 *  @param m	Pointer To System Memory
	 *	@param p	Pointer to Z80's Ports
	 *	@param i	Pointer to Z80's Interrupt Line
     */

	public Z80(Memory m, Ports p, InterruptLine i)
	{
		this.mem = m;
		this.port = p;
		this.irq = i;

		f  = new Flag();
		a  = new Accumulator(f);
		bc = new Registers(f);
		de = new Registers(f);
		hl = new Registers(f);
		ix = new Registers(f);
		iy = new Registers(f);
		r  = new Refresh();
		// Pre-calculate results for DAA instruction
		generateDAATable();
	}


	/**
	 *	Reset Z80.
	 *
	 *  Note that some of these values aren't what a real Z80 would reset to.
	 *  They are the values that the SMS BIOS (to the best of my knowledge)
	 *  sets the registers to.
	 *
	 *	For example, the Index Registers should reset to 0xFFFF
	 *  but doing so breaks 'Prince of Persia', so they are set to 0x0000.
	 *
	 *  The stack pointer is also reset to 0xDFF0 as opposed to 0x0000.
     */

	public final void reset()
	{
		a.clear();
		f.clear();
		bc.clear();
		de.clear();
		hl.clear();
		ix.clear();
		iy.clear();
		r.clear();
		i = 0;

		pc = 0x0000;
		sp = 0xDFF0;
		tstates = 0;

		im = 0;
		iff1 = false;
		iff2 = false;
		EI_inst = false;
		halt = false;
	}


	/**
	 *	Set Accumulator
	 *
	 *	@param reg		Value to set register with
     */

	public final void setA(int reg) {a.set(reg);}
	
	
	/**
	 *	Set Flag Register
	 *
	 *	@param reg		Value to set register with
     */	
	
	public final void setF(int reg) {f.set(reg);}
	
	
	/**
	 *	Set Interrupt Register
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setI(int reg) {i = reg;}
	
	
	/**
	 *	Set Refresh Register
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setR(int reg) {r.set(reg);}
	
	
	/**
	 *	Set BC Register Pair
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setBC(int reg) {bc.set(reg);}
	

	/**
	 *	Set DE Register Pair
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setDE(int reg) {de.set(reg);}
	
	
	/**
	 *	Set HL Register Pair
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setHL(int reg) {hl.set(reg);}
	
	
	/**
	 *	Set IX Register Pair
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setIX(int reg) {ix.set(reg);}
	
	
	/**
	 *	Set IY Register Pair
	 *
	 *	@param reg		Value to set register with
     */
	
	public final void setIY(int reg) {iy.set(reg);}
	
	
	/**
	 *	Set Stack Pointer
	 *
	 *	@param reg		Value to set stack pointer with
     */
	
	public final void setSP(int value) {sp = value;}
	
	
	/**
	 *	Set Interrupt Mode
	 *
	 *	@param reg		Interrupt Mode 0 - 2
     */	
	
	public final void setIM(int mode) {im = mode;}
	
	
	/**
	 *	Set Interrupt Flipflop 2
	 *
	 *	@param flag	True is On, False is Off
     */		
	
	public final void setIFF2(boolean flag) {iff2 = flag;}
	
	
	/**
	 * Exchange registers with shadow registers.
	 */
	
	public final void exall()
	{
		a.ex(); bc.ex(); de.ex(); hl.ex();
	}

	/**
	 *	Get Value of Accumulator
	 *
	 *	@return				Value Stored in Accumulator
     */

	public final int getA() {return a.get();}

	/**
	 *	Get Value of Flag Register
	 *
	 *	@return				Value Stored in Flag Register
     */

	public final int getF() {return f.get();}

	/**
	 *	Get Value of B Register
	 *
	 *	@return				Value Stored in B Register
     */

	public final int getB() {return bc.getH();}

	/**
	 *	Get Value of C Register
	 *
	 *	@return				Value Stored in C Register
     */

	public final int getC() {return bc.getL();}

	/**
	 *	Get Value of D Register
	 *
	 *	@return				Value Stored in D Register
     */

	public final int getD() {return de.getH();}

	/**
	 *	Get Value of E Register
	 *
	 *	@return				Value Stored in E Register
     */

	public final int getE() {return de.getL();}


	/**
	 *	Get Value of H Register
	 *
	 *	@return				Value Stored in H Register
     */

	public final int getH() {return hl.getH();}

	/**
	 *	Get Value of L Register
	 *
	 *	@return				Value Stored in L Register
     */

	public final int getL() {return hl.getL();}

	/**
	 *	Get Value of Refresh Register
	 *
	 *	@return				Value Stored in Refresh Register
     */

	public final int getR() {return r.get();}

	/**
	 *	Get Value of IX Register
	 *
	 *	@return				Value Stored in IX Register
     */

	public final int getIX() {return ix.get();}

	/**
	 *	Get Value of IY Register
	 *
	 *	@return				Value Stored in IY Register
     */

	public final int getIY() {return iy.get();}

	/**
	 *	Get Value of Program Counter
	 *
	 *	@return				Value Stored in Program Counter
     */

	public final int getPC() {return pc;}

	/**
	 *	Get Value of Stack Pointer
	 *
	 *	@return				Value Stored in Stack Pointer
     */

	public final int getSP() {return sp;}

	/**
	 *	Get State of Carry Flag
	 *
	 *	@return				True if set
     */

	public final boolean getCarry() {return f.carry();}

	/**
	 *	Get State of Negative Flag
	 *
	 *	@return				True if set
     */

	public final boolean getNegative() {return f.negative();}

	/**
	 *	Get State of Parity Flag
	 *
	 *	@return				True if set
     */

	public final boolean getParity() {return f.parity();}

	/**
	 *	Get State of Bit 3 Flag
	 *
	 *	@return				True if set
     */

	public final boolean getBit3() {return f.bit3();}

	/**
	 *	Get State of Halfcarry Flag
	 *
	 *	@return				True if set
     */

	public final boolean getHc() {return f.halfcarry();}

	/**
	 *	Get State of Bit 5 Flag
	 *
	 *	@return				True if set
     */

	public final boolean getBit5() {return f.bit5();}

	/**
	 *	Get State of Zero Flag
	 *
	 *	@return				True if set
     */

	public final boolean getZero() {return f.zero();}

	/**
	 *	Get State of Sign Flag
	 *
	 *	@return				True if set
     */

	public final boolean getSign() {return f.sign();}


	/**
	 *	Return Next Opcode for Debugging Purposes
	 *
	 *	@return				String containing opcode bytes
     */

	public final String getOp()
	{
		int opcode = mem.read(pc);
		String oplist = Integer.toHexString(opcode&0xff);

		switch(opcode)
		{
			case 0xCB:
			case 0xED: opcode = mem.read(pc+1); oplist += " "+Integer.toHexString(opcode&0xff); break;
			case 0xDD:
			case 0xFD:
				opcode = mem.read(pc+1);
				oplist += " "+Integer.toHexString(opcode&0xff);
				if (opcode == 0xCB) // DDCB etc
				{
					opcode = mem.read(pc+3);
					oplist += " "+Integer.toHexString(opcode&0xff);
				}
				break;
			default:
				break;
		}

		return oplist.toUpperCase();
	}


	/**
	 *	Return Mnemonic of next opcode for Debugging Purposes
	 *
	 *	@return				String containing opcode bytes
     */

	public final String getMnu()
	{
		int opcode = mem.read(pc);
		Mnemonic mnu = new Mnemonic();
		switch(opcode)
		{
			// special cases
			case 0xDD:
			case 0xFD:
				opcode = mem.read(pc+1); return mnu.getIndex(opcode);
			case 0xCB:
				opcode = mem.read(pc+1); return mnu.getCB(opcode);
			case 0xED:
				opcode = mem.read(pc+1); return mnu.getED(opcode);
			default:
				return mnu.getOP(opcode);
		}
	}


	/**
	 *	Run Z80 for a single instruction (Debugging Purposes)
	 *
	 *  @param cycles		Machine cycles to run for in total
	 *
	 *	@return				True if cycles elapsed
     */

	public final boolean debug(int cycles)
	{
		if (irq.getLine())
			interrupt();

		int opcode = mem.read(pc);		// Fetch Opcode

		doOp(opcode); 					// Interpret Opcode
		tstates += opstates[opcode];	// Increment TStates

		if (tstates > cycles)
		{
			tstates = 0;
			return true;
		}

		return false;
	}


	/**
	 *	Output Contents of Z80 Registers to Console for Debugging Purposes
     */

	private final void consoledebug()
	{
		System.out.println("----------------------------------------------------------------------------");
		System.out.println(Integer.toHexString(pc)+" 0x"+getOp()+" "+getMnu());
		System.out.println("A : "+Integer.toHexString(a.get())+" BC : "+Integer.toHexString(bc.get())+" DE : "+Integer.toHexString(de.get())+" HL : "+Integer.toHexString(hl.get())+
		                   " IX: "+Integer.toHexString(ix.get())+  " IY: "+Integer.toHexString(iy.get()));

		a.ex(); bc.ex(); de.ex(); hl.ex();

		System.out.println("A': "+Integer.toHexString(a.get())+" BC': "+Integer.toHexString(bc.get())+" DE': "+Integer.toHexString(de.get())+ " HL': "+Integer.toHexString(hl.get())+ " SP: "+Integer.toHexString(sp));

		a.ex(); bc.ex(); de.ex(); hl.ex();

		System.out.println("FS: "+getSign()+" FZ: "+getZero()+" FHC: "+getHc()+" FP: "+getParity()+" FN: "+getNegative()+ " FC: "+getCarry());
	}


	/**
	 *	Run Z80
	 *
	 *  @param cycles		Machine cycles to run for in total
	 *
     */

	public final void run(int cycles)
	{
		tstates = 0;
		int opcode = 0;

		while (tstates < cycles)
		{
			if (irq.getLine())
				interrupt(); 					// Check for interrupt

			opcode = mem.read(pc);			   	// Fetch Opcode

			//consoledebug();
			doOp(opcode); 						// Interpret Opcode
			tstates += opstates[opcode];	// Increment TStates
		}
	}


	/**
	 *	Generate Non Maskable Interrupt (NMI)
     */

	public final void nmi()
	{
		iff2 = iff1;
		iff1 = false;

		r.inc();
		// If we're in a halt instruction, increment the PC and get out of it
		if (halt)
		{
			pc++;
			halt = false;
		}

		push(pc);				 // Preserve PC on stack
		pc = 0x66;
		tstates+=11;
	}


	/**
	 *	Normal Interrupt Routine
     */

	public final void interrupt()
	{
		// Interrupts not allowed
		if (!iff1) return;
		// Intterupts not allowed after EI instruction
		if (EI_inst) return;

		// If we're in a halt instruction, increment the PC and get out of it
		if (halt)
		{
			pc++;
			halt = false;
		}

		r.inc();
		iff1 = false;
		iff2 = false;
		irq.setLine(false);

		if ((im == 1) || (im == 0))
		{
			push(pc);				 // Preserve PC on stack
			pc = 0x38;
			tstates+=13;
		}
		else
			System.out.println("Invalid IM mode "+im);
	}


	/**
	 *	Execute Opcode
	 *
	 *	@param opcode		Opcode hex value
     */

	private final void doOp(int opcode)
	{
			int temp;

			r.inc();														// Increase Memory Refresh Register
			EI_inst = false;

			switch(opcode)
			{
				case 0x00: pc++; break;										// NOP
				case 0x01: bc.set(mem.readWord(pc+1)); pc+=3; break;			// LD BC,nn
				case 0x02: mem.write(bc.get(), a.get()); pc++; break; 		// LD (BC),A
				case 0x03: bc.inc(); pc++; break; 							// INC BC
				case 0x04: bc.incH(); pc++; break; 							// INC B
				case 0x05: bc.decH(); pc++; break;							// DEC B
				case 0x06: bc.setH(mem.read(pc+1)); pc+=2; break; 			// LD B,n
				case 0x07: a.rlca(); pc++; break; 							// RLCA
				case 0x08: a.ex(); f.ex(); pc++; break; 						// EX AF AF'
				case 0x09: hl.add(bc.get()); pc++; break; 					// ADD HL,BC
				case 0x0A: a.set(mem.read(bc.get())); pc++; break; 			// LD A,(BC)
				case 0x0B: bc.dec(); pc++; break; 							// DEC BC
				case 0x0C: bc.incL(); pc++; break; 							// INC C
				case 0x0D: bc.decL(); pc++; break;							// DEC C
				case 0x0E: bc.setL(mem.read(pc+1)); pc+=2; break; 			// LD C,n
				case 0x0F: a.rrca(); pc++; break; 							// RRCA
				case 0x10: temp = (bc.getH()-1)&0xff;							// DJNZ (PC+e)
						   jr(temp != 0); bc.setH(temp); break;
				case 0x11: de.set(mem.readWord(pc+1)); pc+=3; break;			// LD DE,nn
				case 0x12: mem.write(de.get(), a.get()); pc++; break;			// LD (DE), A
				case 0x13: de.inc(); pc++; break;								// INC DE
				case 0x14: de.incH(); pc++; break; 							// INC D
				case 0x15: de.decH(); pc++; break;							// DEC D
				case 0x16: de.setH(mem.read(pc+1)); pc+=2; break; 			// LD D,n
				case 0x17: a.rla(); pc++; break; 								// RLA
				case 0x18: pc+= mem.readSigned(pc+1) +2; break; 				// JR (PC+e)
				case 0x19: hl.add(de.get()); pc++; break; 					// ADD HL,DE
				case 0x1A: a.set(mem.read(de.get())); pc++; break; 			// LD A,(DE)
				case 0x1B: de.dec(); pc++; break; 							// DEC DE
				case 0x1C: de.incL(); pc++; break; 							// INC E
				case 0x1D: de.decL(); pc++; break; 							// DEC E
				case 0x1E: de.setL(mem.read(pc+1)); pc+=2; break; 			// LD E,N
				case 0x1F: a.rra(); pc++; break; 								// RRA (c)
				case 0x20: jr(!f.zero()); break; 								// JR NZ,(PC+e)
				case 0x21: hl.set(mem.readWord(pc+1)); pc+=3; break; 			// LD HL,nn
				case 0x22: mem.write(mem.readWord(pc+1), hl.getL());			// LD (nn),HL
						   mem.write((mem.readWord(pc+1)+1), hl.getH());
						   pc+=3; break;
				case 0x23: hl.inc(); pc++; break; 							// INC HL
				case 0x24: hl.incH(); pc++; break; 							// INC H
				case 0x25: hl.decH(); pc++; break; 							// DEC H
				case 0x26: hl.setH(mem.read(pc+1)); pc+=2; break; 			// LD H,n
				case 0x27: daa(); pc++; break;								// DAA
				case 0x28: jr(f.zero()); break;								// JR Z,(PC+e)
				case 0x29: hl.add(hl.get()); pc++; break; 					// ADD HL,HL
				case 0x2A: hl.setL(mem.read(mem.readWord(pc+1)));				// LD HL,(nn)
						   hl.setH(mem.read(mem.readWord(pc+1)+1));
						   pc+=3; break;
				case 0x2B: hl.dec(); pc++; break; 							// DEC HL
				case 0x2C: hl.incL(); pc++; break; 							// INC L
				case 0x2D: hl.decL(); pc++; break; 							// DEC L
				case 0x2E: hl.setL(mem.read(pc+1)); pc+=2; break; 			// LD L,n
				case 0x2F: a.cpl(); pc++; break;								// CPL
				case 0x30: jr(!f.carry()); break;								// JR NC,(PC+e)
				case 0x31: sp = mem.readWord(pc+1); pc+=3; break; 			// LD SP,nn
				case 0x32: mem.write(mem.readWord(pc+1), a.get()); pc+=3; break; // LD (nn),A
				case 0x33: sp++; pc++; break; 								// INC SP
				case 0x34: inchl(hl.get()); pc++; break;						// INC (HL)
				case 0x35: dechl(hl.get()); pc++; break;						// DEC (HL)
				case 0x36: mem.write(hl.get(), mem.read(pc+1)); pc+=2;break;	// LD (HL),n
				case 0x37: f.carryOn(); f.negativeOff(); f.halfCarryOff();;	// SCF
						   pc++; break;
				case 0x38: jr(f.carry()); break;								// JR C,(PC+e)
				case 0x39: hl.add(sp); pc++; break; 							// ADD HL,SP
				case 0x3A: a.set(mem.read(mem.readWord(pc+1))); pc+=3; break;	// LD A,(nn)
				case 0x3B: sp--; pc++; break; 								// DEC SP
				case 0x3C: a.inc(); pc++; break; 								// INC A
				case 0x3D: a.dec(); pc++; break; 								// DEC A
				case 0x3E: a.set(mem.read(pc+1)); pc+=2; break; 				// LD A,n
				case 0x3F: ccf(); pc++; break;								// CCF
				case 0x40: 									pc++; break; 	// LD B,B
				case 0x41: bc.setH(bc.getL());      		pc++; break;		// LD B,C
				case 0x42: bc.setH(de.getH());     			pc++; break;	// LD B,D
				case 0x43: bc.setH(de.getL());      		pc++; break; 		// LD B,E
				case 0x44: bc.setH(hl.getH());     			pc++; break; 	// LD B,H
				case 0x45: bc.setH(hl.getL());      		pc++; break; 		// LD B,L
				case 0x46: bc.setH(mem.read(hl.get())); 	pc++; break; 		// LD B,(HL)
				case 0x47: bc.setH(a.get());          		pc++; break; 		// LD B,A
				case 0x48: bc.setL(bc.getH());				pc++; break; 		// LD C,B
				case 0x49: 									pc++; break; 	// LD C,C
				case 0x4A: bc.setL(de.getH());     			pc++; break;	// LD C,D
				case 0x4B: bc.setL(de.getL());      		pc++; break; 		// LD C,E
				case 0x4C: bc.setL(hl.getH());     			pc++; break; 	// LD C,H
				case 0x4D: bc.setL(hl.getL());      		pc++; break; 		// LD C,L
				case 0x4E: bc.setL(mem.read(hl.get())); 	pc++; break; 		// LD C,(HL)
				case 0x4F: bc.setL(a.get());          		pc++; break; 		// LD C,A
				case 0x50: de.setH(bc.getH());				pc++; break; 		// LD D,B
				case 0x51: de.setH(bc.getL());      		pc++; break;		// LD D,C
				case 0x52:      							pc++; break;		// LD D,D
				case 0x53: de.setH(de.getL());      		pc++; break; 		// LD D,E
				case 0x54: de.setH(hl.getH());     			pc++; break; 	// LD D,H
				case 0x55: de.setH(hl.getL());      		pc++; break; 		// LD D,L
				case 0x56: de.setH(mem.read(hl.get())); 	pc++; break; 		// LD D,(HL)
				case 0x57: de.setH(a.get());          		pc++; break; 		// LD D,A
				case 0x58: de.setL(bc.getH());				pc++; break; 		// LD E,B
				case 0x59: de.setL(bc.getL());      		pc++; break;		// LD E,C
				case 0x5A: de.setL(de.getH());     			pc++; break;	// LD E,D
				case 0x5B:       							pc++; break; 		// LD E,E
				case 0x5C: de.setL(hl.getH());     			pc++; break; 	// LD E,H
				case 0x5D: de.setL(hl.getL());      		pc++; break; 		// LD E,L
				case 0x5E: de.setL(mem.read(hl.get())); 	pc++; break; 		// LD E,(HL)
				case 0x5F: de.setL(a.get());          		pc++; break; 		// LD E,A
				case 0x60: hl.setH(bc.getH());				pc++; break; 		// LD H,B
				case 0x61: hl.setH(bc.getL());      		pc++; break;		// LD H,C
				case 0x62: hl.setH(de.getH());     			pc++; break;	// LD H,D
				case 0x63: hl.setH(de.getL());      		pc++; break; 		// LD H,E
				case 0x64:    								pc++; break; 		// LD H,H
				case 0x65: hl.setH(hl.getL());      		pc++; break; 		// LD H,L
				case 0x66: hl.setH(mem.read(hl.get())); 	pc++; break; 		// LD H,(HL)
				case 0x67: hl.setH(a.get());          		pc++; break; 		// LD H,A
				case 0x68: hl.setL(bc.getH());				pc++; break; 		// LD L,B
				case 0x69: hl.setL(bc.getL());      		pc++; break;		// LD L,C
				case 0x6A: hl.setL(de.getH());     			pc++; break;	// LD L,D
				case 0x6B: hl.setL(de.getL());      		pc++; break; 		// LD L,E
				case 0x6C: hl.setL(hl.getH());     			pc++; break; 	// LD L,H
				case 0x6D: 						      		pc++; break; 	// LD L,L
				case 0x6E: hl.setL(mem.read(hl.get())); 	pc++; break; 		// LD L,(HL)
				case 0x6F: hl.setL(a.get());        		pc++; break; 		// LD L,A
				case 0x70: mem.write(hl.get(), bc.getH());	pc++; break; 		// LD (HL),B
				case 0x71: mem.write(hl.get(), bc.getL());	pc++; break; 		// LD (HL),C
				case 0x72: mem.write(hl.get(), de.getH());	pc++; break; 		// LD (HL),D
				case 0x73: mem.write(hl.get(), de.getL());	pc++; break; 		// LD (HL),E
				case 0x74: mem.write(hl.get(), hl.getH());	pc++; break; 		// LD (HL),H
				case 0x75: mem.write(hl.get(), hl.getL()); 	pc++; break; 	// LD (HL),L
				case 0x76: halt = true; 						   break; 		// HALT
				case 0x77: mem.write(hl.get(), a.get());	pc++; break; 		// LD (HL),A
				case 0x78: a.set(bc.getH());				pc++; break; 		// LD A,B
				case 0x79: a.set(bc.getL());				pc++; break; 		// LD A,C
				case 0x7A: a.set(de.getH());				pc++; break; 		// LD A,D
				case 0x7B: a.set(de.getL());				pc++; break; 		// LD A,E
				case 0x7C: a.set(hl.getH());				pc++; break; 		// LD A,H
				case 0x7D: a.set(hl.getL());				pc++; break; 		// LD A,L
				case 0x7E: a.set(mem.read(hl.get()));		pc++; break; 		// LD A,(HL)
				case 0x7F:									pc++; break; 		// LD A,A
				case 0x80: a.add(bc.getH());				pc++; break; 		// ADD A,B
				case 0x81: a.add(bc.getL());				pc++; break; 		// ADD A,C
				case 0x82: a.add(de.getH());				pc++; break; 		// ADD A,D
				case 0x83: a.add(de.getL());				pc++; break; 		// ADD A,E
				case 0x84: a.add(hl.getH());				pc++; break;	 	// ADD A,H
				case 0x85: a.add(hl.getL());				pc++; break; 		// ADD A,L
				case 0x86: a.add(mem.read(hl.get()));		pc++; break; 		// ADD A,(HL)
				case 0x87: a.add(a.get());					pc++; break; 		// ADD A,A
				case 0x88: a.adc(bc.getH());				pc++; break; 		// ADC A,B
				case 0x89: a.adc(bc.getL());				pc++; break; 		// ADC A,C
				case 0x8A: a.adc(de.getH());				pc++; break; 		// ADC A,D
				case 0x8B: a.adc(de.getL());				pc++; break; 		// ADC A,E
				case 0x8C: a.adc(hl.getH());				pc++; break; 		// ADC A,H
				case 0x8D: a.adc(hl.getL());				pc++; break; 		// ADC A,L
				case 0x8E: a.adc(mem.read(hl.get()));		pc++; break; 		// ADC A,(HL)
				case 0x8F: a.adc(a.get());					pc++; break; 		// ADC A,A
				case 0x90: a.sub(bc.getH());				pc++; break; 		// SUB A,B
				case 0x91: a.sub(bc.getL());				pc++; break; 		// SUB A,C
				case 0x92: a.sub(de.getH());				pc++; break; 		// SUB A,D
				case 0x93: a.sub(de.getL());				pc++; break; 		// SUB A,E
				case 0x94: a.sub(hl.getH());				pc++; break; 		// SUB A,H
				case 0x95: a.sub(hl.getL());				pc++; break; 		// SUB A,L
				case 0x96: a.sub(mem.read(hl.get()));		pc++; break; 		// SUB A,(HL)
				case 0x97: a.sub(a.get());					pc++; break; 		// SUB A,A
				case 0x98: a.sbc(bc.getH());				pc++; break; 		// SBC A,B
				case 0x99: a.sbc(bc.getL());				pc++; break; 		// SBC A,C
				case 0x9A: a.sbc(de.getH());				pc++; break; 		// SBC A,D
				case 0x9B: a.sbc(de.getL());				pc++; break; 		// SBC A,E
				case 0x9C: a.sbc(hl.getH());				pc++; break; 		// SBC A,H
				case 0x9D: a.sbc(hl.getL());				pc++; break; 		// SBC A,L
				case 0x9E: a.sbc(mem.read(hl.get()));		pc++; break; 		// SBC A,(HL)
				case 0x9F: a.sbc(a.get());					pc++; break; 		// SBC A,A
				case 0xA0: a.and(bc.getH());				pc++; break; 		// AND A,B
				case 0xA1: a.and(bc.getL());				pc++; break; 		// AND A,C
				case 0xA2: a.and(de.getH());				pc++; break; 		// AND A,D
				case 0xA3: a.and(de.getL());				pc++; break; 		// AND A,E
				case 0xA4: a.and(hl.getH());				pc++; break; 		// AND A,H
				case 0xA5: a.and(hl.getL());				pc++; break; 		// AND A,L
				case 0xA6: a.and(mem.read(hl.get()));		pc++; break; 		// AND A,(HL)
				case 0xA7: a.and(a.get());					pc++; break; 		// AND A,A
				case 0xA8: a.xor(bc.getH());				pc++; break; 		// XOR A,B
				case 0xA9: a.xor(bc.getL());				pc++; break; 		// XOR A,C
				case 0xAA: a.xor(de.getH());				pc++; break; 		// XOR A,D
				case 0xAB: a.xor(de.getL());				pc++; break; 		// XOR A,E
				case 0xAC: a.xor(hl.getH());				pc++; break; 		// XOR A,H
				case 0xAD: a.xor(hl.getL());				pc++; break; 		// XOR A,L
				case 0xAE: a.xor(mem.read(hl.get()));		pc++; break; 		// XOR A,(HL)
				case 0xAF: a.xor(a.get());					pc++; break; 		// XOR A,A (=0)
				case 0xB0: a.or(bc.getH());					pc++; break; 	// OR A,B
				case 0xB1: a.or(bc.getL());					pc++; break; 	// OR A,C
				case 0xB2: a.or(de.getH());					pc++; break; 	// OR A,D
				case 0xB3: a.or(de.getL());					pc++; break; 	// OR A,E
				case 0xB4: a.or(hl.getH());					pc++; break; 	// OR A,H
				case 0xB5: a.or(hl.getL());					pc++; break; 	// OR A,L
				case 0xB6: a.or(mem.read(hl.get()));		pc++; break; 		// OR A,(HL)
				case 0xB7: a.or(a.get());					pc++; break; 		// OR A,A
				case 0xB8: a.cp(bc.getH());					pc++; break; 	// CP A,B
				case 0xB9: a.cp(bc.getL());					pc++; break; 	// CP A,C
				case 0xBA: a.cp(de.getH());					pc++; break; 	// CP A,D
				case 0xBB: a.cp(de.getL());					pc++; break; 	// CP A,E
				case 0xBC: a.cp(hl.getH());					pc++; break; 	// CP A,H
				case 0xBD: a.cp(hl.getL());					pc++; break; 	// CP A,L
				case 0xBE: a.cp(mem.read(hl.get()));		pc++; break; 		// CP A,(HL)
				case 0xBF: a.cp(a.get());					pc++; break; 		// CP A,A
				case 0xC0: ret(!f.zero());						  break;		// RET NZ
				case 0xC1: bc.set(pop());					pc++; break; 		// POP BC
				case 0xC2: jp(!f.zero());						  break;		// JP NZ,(nn)
				case 0xC3: pc = mem.readWord(pc+1);				  break; 	// JP (nn)
				case 0xC4: call(!f.zero());						  break;	// CALL NZ (nn)
				case 0xC5: push(bc.get());					pc++; break;		// PUSH BC
				case 0xC6: a.add(mem.read(++pc));			pc++; break; 		// ADD A,n
				case 0xC7: push(pc+1); pc=0x00;					  break;	// RST 00H
				case 0xC8: ret(f.zero());						  break;		// RET Z
				case 0xC9: pc = pop();		  					  break;		// RET
				case 0xCA: jp(f.zero());						  break;		// JP Z,(nn)
				case 0xCB: doCB(mem.read(++pc)); 		  		  break;		// CB Opcode
				case 0xCC: call(f.zero());						  break;		// CALL Z (nn)
				case 0xCD: push(pc+3);pc = mem.readWord(pc+1);	  break;		// CALL (nn)
				case 0xCE: a.adc(mem.read(++pc));			pc++; break; 		// ADC A,n
				case 0xCF: push(pc+1); pc=0x08;					  break;	// RST 08H
				case 0xD0: ret(!f.carry());						  break;	// RET NC
				case 0xD1: de.set(pop());					pc++; break; 		// POP DE
				case 0xD2: jp(!f.carry());						  break;		// JP NC,(nn)
				case 0xD3: port.out(mem.read(++pc), a.get());pc++;break; 		// OUT (n),A
				case 0xD4: call(!f.carry());					  break;		// CALL NC (nn)
				case 0xD5: push(de.get());					pc++; break;		// PUSH DE
				case 0xD6: a.sub(mem.read(++pc));			pc++; break; 		// SUB n
				case 0xD7: push(pc+1); pc=0x10;					  break;	// RST 10H
				case 0xD8: ret(f.carry());						  break;		// RET C
				case 0xD9: bc.ex(); de.ex(); hl.ex();		pc++; break; 		// EXX
				case 0xDA: jp(f.carry());						  break;		// JP C,(nn)
				case 0xDB: a.set(port.in(mem.read(++pc)));	pc++; break; 		// IN A,(n)
				case 0xDC: call(f.carry());						  break;	// CALL C (nn)
				case 0xDD: doIndexOp(mem.read(++pc), ix);		  break;		// DD Opcode
				case 0xDE: a.sbc(mem.read(++pc)); pc++;			  break; 	// SBC A,n
				case 0xDF: push(pc+1); pc=0x18;					  break;	// RST 18H
				case 0xE0: ret(!f.parity());					  break;		// RET PO
				case 0xE1: hl.set(pop());					pc++; break; 		// POP HL
				case 0xE2: jp(!f.parity());						  break;	// JP PO,(nn)
				case 0xE3: temp = hl.get(); hl.set(mem.readWord(sp));			// EX (SP),HL
						   mem.write(sp, temp&0xff);
						   mem.write(sp+1, temp>>8);		pc++; break;
				case 0xE4: call(!f.parity());					  break;		// CALL PO (nn)
				case 0xE5: push(hl.get());					pc++; break;		// PUSH HL
				case 0xE6: a.and(mem.read(++pc));			pc++; break; 		// AND (n)
				case 0xE7: push(pc+1); pc=0x20;					break;		// RST 20H
				case 0xE8: ret(f.parity());						  break;	// RET PE
				case 0xE9: pc = hl.get();					  	  break; 		// JP (HL)
				case 0xEA: jp(f.parity());						  break;		// JP PE,(nn)
				case 0xEB: temp = de.get(); de.set(hl.get());					// EX DE,HL
						   hl.set(temp);					pc++; break;
				case 0xEC: call(f.parity());					  break;		// CALL PE (nn)
				case 0xED: doED(mem.read(++pc)); 				  break;		// ED Opcode
				case 0xEE: a.xor(mem.read(++pc));			pc++; break; 		// XOR n
				case 0xEF: push(pc+1); pc=0x28;					  break;	// RST 28H
				case 0xF0: ret(!f.sign());						  break;		// RET P
				case 0xF1: f.set(mem.read(sp)); a.set(mem.read(sp+1));			// POP AF
						   sp+=2;							pc++; break;
				case 0xF2: jp(!f.sign());						  break;		// JP P,(nn)
				case 0xF3: iff1=false;iff2=false;EI_inst=true;pc++; break; 	// DI
				case 0xF4: call(!f.sign());					  	  break;	// CALL P (nn)
				case 0xF5: push((a.get()<<8)+f.get());		pc++; break;		// PUSH AF
				case 0xF6: a.or(mem.read(++pc));			pc++; break; 		// OR n
				case 0xF7: push(pc+1); pc=0x30;					  break;	// RST 30H
				case 0xF8: ret(f.sign());						  break;		// RET M
				case 0xF9: sp = hl.get();					pc++; break; 		// LD SP,HL
				case 0xFA: jp(f.sign());				    	  break;		// JP M,(nn)
				case 0xFB: iff1=true;iff2=true;EI_inst=true; pc++;break;		// EI
				case 0xFC: call(f.sign());						  break;		// CALL M (nn)
				case 0xFD: doIndexOp(mem.read(++pc), iy);		  break;		// FD Opcode
				case 0xFE: a.cp(mem.read(++pc)); 			pc++; break; 		// CP n
				case 0xFF: push(pc+1); pc=0x38;				break;			// RST 38H

				// Unimplemented Opcode
				default:
					System.out.println("Unimplemented Opcode: "+Integer.toHexString(opcode&0xff));
					pc++;
					break;

			} // end switch
	}


	/**
	 *	Jump
	 *
	 *  @param condition		If true jump will be taken
     */

	private final void jp(boolean condition)
	{
		if (condition) pc = mem.readWord(pc+1);
		else pc+=3;
	}


	/**
	 *	Jump Relative
	 *
	 *  @param condition		If true jump will be taken
     */

	private final void jr(boolean condition)
	{
		if (condition)
		{
			pc+= mem.readSigned(pc+1) +2;
			tstates+=5;
		}
		else pc+=2;
	}


	/**
	 *	Call
	 *
	 *  @param condition		If true call will be taken
     */

	private final void call(boolean condition)
	{
		if (condition)
		{
			push(pc+3);					// write value of PC to stack
			pc = mem.readWord(pc+1);
			tstates+=7;
		}
		else pc+=3;
	}


	/**
	 *	Return
	 *
	 *  @param condition		If true return will be taken
     */

	private final void ret(boolean condition)
	{
		if (condition)
		{
			pc = pop();
			tstates+=6;
		}
		else pc++;
	}


	/**
	 *	Push Value Onto Stack
	 *
	 *  @param value		Value to push
     */

	private final void push(int value)
	{
		mem.write(--sp, value>>8);		// (SP - 1) <- high
		mem.write(--sp, value&0xff);	// (SP - 2) <- low
	}
	
	/**
	 *	Push Value Off Stack
	 *
	 *  @return	Value from stack
     */

	private final int pop()
	{
		int value = mem.readWord(sp); 
		sp+=2;
		sp &= 0xFFFF;
		return value;
	}


	/**
	 *	INC (HL) - Increment Memory Location
	 *
	 *  @param offset		Memory Offset to Increment
     */

	private final void inchl(int offset)
	{
		offset &= 0xFFFF;
		int location = mem.read(offset);
		int result = location + 1;
		result &= 0xFF;

		if ((result & 0x0F) == 0x00) f.halfCarryOn();
		else f.halfCarryOff();
		if (result == 0x80) f.parityOn();
		else f.parityOff();
		if (result == 0) f.zeroOn();
		else f.zeroOff();
		if ((result & 0x80) == 0) f.signOff();
		else f.signOn();
		f.negativeOff();
		f.setBit3(result);
		f.setBit5(result);

		mem.write(offset, result);
	}


	/**
	 *	DEC (HL) - Decrement Memory Location
	 *
	 *  @param offset		Memory Offset to Increment
     */

	private final void dechl(int offset)
	{
		offset &= 0xFFFF;
		int location = mem.read(offset);
		int result = location - 1;
		result &= 0xFF;


		if ((result & 0x0F) == 0x0F) f.halfCarryOn();
		else f.halfCarryOff();
		if (result == 0x7F) f.parityOn();
		else f.parityOff();
		if (result == 0) f.zeroOn();
		else f.zeroOff();
		if ((result & 0x80) == 0) f.signOff();
		else f.signOn();
		f.negativeOn();
		f.setBit3(result);
		f.setBit5(result);

		mem.write(offset, result);
	}

	/**
	 *	CCF - Complement Carry Flag
     */

	private final void ccf()
	{
		if (f.carry())
		{
			f.carryOff();
			f.halfCarryOn();
		}
		else
		{
			f.carryOn();
			f.halfCarryOff();
		}
		f.negativeOff();
	}


	/**
	 *	DAA - Decimal Adjust Accumulator
	 *  adds 6 to left and/or right nibble
	 *
	 *  Pre-Calculated Result For Speed
     */

	private final void daa()
	{
		int a_copy = a.get();

		// Mask on extra bits
		if (f.negative())
			a_copy |= 0x100;
		if (f.carry())
			a_copy |= 0x200;
		if (f.halfcarry())
			a_copy |= 0x400;
		// Get result for calculated table
		int temp = daa[a_copy];

		if (temp > 0xFF) f.carryOn();
		else f.carryOff();

		temp &= 0xFF;
		a.set(temp);

		f.setParity(temp);

		if (temp == 0) f.zeroOn();
		else f.zeroOff();
		if ((temp & 0x80) == 0) f.signOff();
		else f.signOn();
	}


	/**
	 *	Execute CB Prefixed Opcode
	 *
	 *	@param opcode		Opcode hex value
     */

	private final void doCB(int opcode)
	{
		r.inc();				 												// Increment Memory Refresh
		tstates += opcbstates[opcode];

		switch(opcode)
		{
			case 0x00: bc.setH(rlc(bc.getH())); break; 							// RLC B
			case 0x01: bc.setL(rlc(bc.getL())); break;							// RLC C
			case 0x02: de.setH(rlc(de.getH())); break; 							// RLC D
			case 0x03: de.setL(rlc(de.getL())); break;							// RLC E
			case 0x04: hl.setH(rlc(hl.getH())); break; 							// RLC H
			case 0x05: hl.setL(rlc(hl.getL())); break;							// RLC L
			case 0x06: mem.write(hl.get(), rlc(mem.read(hl.get()))); break;		// RLC (HL)
			case 0x07: a.set(rlc(a.get())); break;								// RLC A
			case 0x08: bc.setH(rrc(bc.getH())); break; 							// RRC B
			case 0x09: bc.setL(rrc(bc.getL())); break;							// RRC C
			case 0x0A: de.setH(rrc(de.getH())); break; 							// RRC D
			case 0x0B: de.setL(rrc(de.getL())); break;							// RRC E
			case 0x0C: hl.setH(rrc(hl.getH())); break; 							// RRC H
			case 0x0D: hl.setL(rrc(hl.getL())); break;							// RRC L
			case 0x0E: mem.write(hl.get(), rrc(mem.read(hl.get())));	break;		// RRC (HL)
			case 0x0F: a.set(rrc(a.get())); break;								// RRC A
			case 0x10: bc.setH(rl(bc.getH())); break; 							// RL B
			case 0x11: bc.setL(rl(bc.getL())); break;								// RL C
			case 0x12: de.setH(rl(de.getH())); break; 							// RL D
			case 0x13: de.setL(rl(de.getL())); break;								// RL E
			case 0x14: hl.setH(rl(hl.getH())); break; 							// RL H
			case 0x15: hl.setL(rl(hl.getL())); break;								// RL L
			case 0x16: mem.write(hl.get(), rl(mem.read(hl.get()))); break;		// RL (HL)
			case 0x17: a.set(rl(a.get())); break;									// RL A
			case 0x18: bc.setH(rr(bc.getH())); break; 							// RR B
			case 0x19: bc.setL(rr(bc.getL())); break;								// RR C
			case 0x1A: de.setH(rr(de.getH())); break; 							// RR D
			case 0x1B: de.setL(rr(de.getL())); break;								// RR E
			case 0x1C: hl.setH(rr(hl.getH())); break; 							// RR H
			case 0x1D: hl.setL(rr(hl.getL())); break;								// RR L
			case 0x1E: mem.write(hl.get(), rr(mem.read(hl.get()))); break;		// RR (HL)
			case 0x1F: a.set(rr(a.get())); break;									// RR A
			case 0x20: bc.setH(sla(bc.getH())); break; 							// SLA B
			case 0x21: bc.setL(sla(bc.getL())); break;							// SLA C
			case 0x22: de.setH(sla(de.getH())); break; 							// SLA D
			case 0x23: de.setL(sla(de.getL())); break;							// SLA E
			case 0x24: hl.setH(sla(hl.getH())); break; 							// SLA H
			case 0x25: hl.setL(sla(hl.getL())); break;							// SLA L
			case 0x26: mem.write(hl.get(), sla(mem.read(hl.get()))); break;		// SLA (HL)
			case 0x27: a.set(sla(a.get())); break;								// SLA A
			case 0x28: bc.setH(sra(bc.getH())); break; 							// SRA B
			case 0x29: bc.setL(sra(bc.getL())); break;							// SRA C
			case 0x2A: de.setH(sra(de.getH())); break; 							// SRA D
			case 0x2B: de.setL(sra(de.getL())); break;							// SRA E
			case 0x2C: hl.setH(sra(hl.getH())); break; 							// SRA H
			case 0x2D: hl.setL(sra(hl.getL())); break;							// SRA L
			case 0x2E: mem.write(hl.get(), sra(mem.read(hl.get()))); break;		// SRA (HL)
			case 0x2F: a.set(sra(a.get())); break;								// SRA A
			case 0x30: bc.setH(sll(bc.getH())); break; 							// SLL B
			case 0x31: bc.setL(sll(bc.getL())); break;							// SLL C
			case 0x32: de.setH(sll(de.getH())); break; 							// SLL D
			case 0x33: de.setL(sll(de.getL())); break;							// SLL E
			case 0x34: hl.setH(sll(hl.getH())); break; 							// SLL H
			case 0x35: hl.setL(sll(hl.getL())); break;							// SLL L
			case 0x36: mem.write(hl.get(), sll(mem.read(hl.get()))); break;		// SLL (HL)	
			case 0x37: a.set(sll(a.get())); break;								// SLL A
			case 0x38: bc.setH(srl(bc.getH())); break; 							// SRL B
			case 0x39: bc.setL(srl(bc.getL())); break;							// SRL C
			case 0x3A: de.setH(srl(de.getH())); break; 							// SRL D
			case 0x3B: de.setL(srl(de.getL())); break;							// SRL E
			case 0x3C: hl.setH(srl(hl.getH())); break; 							// SRL H
			case 0x3D: hl.setL(srl(hl.getL())); break;							// SRL L
			case 0x3E: mem.write(hl.get(), srl(mem.read(hl.get()))); break;		// SRL (HL)
			case 0x3F: a.set(srl(a.get())); break;								// SRL A
			case 0x40: bit(bc.getH(), 0); break;									// BIT 0,B
			case 0x41: bit(bc.getL(),  0); break;									// BIT 0,C
			case 0x42: bit(de.getH(), 0); break;									// BIT 0,D
			case 0x43: bit(de.getL(),  0); break;									// BIT 0,E
			case 0x44: bit(hl.getH(), 0); break;									// BIT 0,H
			case 0x45: bit(hl.getL(),  0); break;									// BIT 0,L
			case 0x46: bit(mem.read(hl.get()), 0); break;							// BIT 0,(HL)
			case 0x47: bit(a.get(), 0); break;									// BIT 0,A
			case 0x48: bit(bc.getH(), 1); break;									// BIT 1,B
			case 0x49: bit(bc.getL(),  1); break;									// BIT 1,C
			case 0x4A: bit(de.getH(), 1); break;									// BIT 1,D
			case 0x4B: bit(de.getL(),  1); break;									// BIT 1,E
			case 0x4C: bit(hl.getH(), 1); break;									// BIT 1,H
			case 0x4D: bit(hl.getL(),  1); break;									// BIT 1,L
			case 0x4E: bit(mem.read(hl.get()), 1); break;							// BIT 1,(HL)
			case 0x4F: bit(a.get(), 1); break;									// BIT 1,A
			case 0x50: bit(bc.getH(), 2); break;									// BIT 2,B
			case 0x51: bit(bc.getL(),  2); break;									// BIT 2,C
			case 0x52: bit(de.getH(), 2); break;									// BIT 2,D
			case 0x53: bit(de.getL(),  2); break;									// BIT 2,E
			case 0x54: bit(hl.getH(), 2); break;									// BIT 2,H
			case 0x55: bit(hl.getL(),  2); break;									// BIT 2,L
			case 0x56: bit(mem.read(hl.get()), 2); break;							// BIT 2,(HL)
			case 0x57: bit(a.get(), 2); break;									// BIT 2,A
			case 0x58: bit(bc.getH(), 3); break;									// BIT 3,B
			case 0x59: bit(bc.getL(),  3); break;									// BIT 3,C
			case 0x5A: bit(de.getH(), 3); break;									// BIT 3,D
			case 0x5B: bit(de.getL(),  3); break;									// BIT 3,E
			case 0x5C: bit(hl.getH(), 3); break;									// BIT 3,H
			case 0x5D: bit(hl.getL(),  3); break;									// BIT 3,L
			case 0x5E: bit(mem.read(hl.get()), 3); break;							// BIT 3,(HL)
			case 0x5F: bit(a.get(), 3); break;									// BIT 3,A
			case 0x60: bit(bc.getH(), 4); break;									// BIT 4,B
			case 0x61: bit(bc.getL(),  4); break;									// BIT 4,C
			case 0x62: bit(de.getH(), 4); break;									// BIT 4,D
			case 0x63: bit(de.getL(),  4); break;									// BIT 4,E
			case 0x64: bit(hl.getH(), 4); break;									// BIT 4,H
			case 0x65: bit(hl.getL(),  4); break;									// BIT 4,L
			case 0x66: bit(mem.read(hl.get()), 4); break;							// BIT 4,(HL)
			case 0x67: bit(a.get(), 4); break;									// BIT 4,A
			case 0x68: bit(bc.getH(), 5); break;									// BIT 5,B
			case 0x69: bit(bc.getL(),  5); break;									// BIT 5,C
			case 0x6A: bit(de.getH(), 5); break;									// BIT 5,D
			case 0x6B: bit(de.getL(),  5); break;									// BIT 5,E
			case 0x6C: bit(hl.getH(), 5); break;									// BIT 5,H
			case 0x6D: bit(hl.getL(),  5); break;									// BIT 5,L
			case 0x6E: bit(mem.read(hl.get()), 5); break;							// BIT 5,(HL)
			case 0x6F: bit(a.get(), 		5); break;								// BIT 5,A
			case 0x70: bit(bc.getH(), 6); break;									// BIT 6,B
			case 0x71: bit(bc.getL(),  6); break;									// BIT 6,C
			case 0x72: bit(de.getH(), 6); break;									// BIT 6,D
			case 0x73: bit(de.getL(),  6); break;									// BIT 6,E
			case 0x74: bit(hl.getH(), 6); break;									// BIT 6,H
			case 0x75: bit(hl.getL(),  6); break;									// BIT 6,L
			case 0x76: bit(mem.read(hl.get()), 6); break;							// BIT 6,(HL)
			case 0x77: bit(a.get(), 		6); break;								// BIT 6,A
			case 0x78: bit(bc.getH(), 7); break;									// BIT 7,B
			case 0x79: bit(bc.getL(),  7); break;									// BIT 7,C
			case 0x7A: bit(de.getH(), 7); break;									// BIT 7,D
			case 0x7B: bit(de.getL(),  7); break;									// BIT 7,E
			case 0x7C: bit(hl.getH(), 7); break;									// BIT 7,H
			case 0x7D: bit(hl.getL(),  7); break;									// BIT 7,L
			case 0x7E: bit(mem.read(hl.get()), 7); break;							// BIT 7,(HL)
			case 0x7F: bit(a.get(), 7); break;									// BIT 7,A
			case 0x80: bc.setH(res(bc.getH(), 0)); break;							// RES 0,B
			case 0x81: bc.setL(res(bc.getL(),   0)); break;						// RES 0,C
			case 0x82: de.setH(res(de.getH(), 0)); break;							// RES 0,D
			case 0x83: de.setL(res(de.getL(),   0)); break;						// RES 0,E
			case 0x84: hl.setH(res(hl.getH(), 0)); break;							// RES 0,H
			case 0x85: hl.setL(res(hl.getL(),   0)); break;						// RES 0,L
			case 0x86: mem.write(hl.get(), res(mem.read(hl.get()), 0)); break;	// RES 0,(HL)
			case 0x87: a.set(res(a.get(),           0)); break;					// RES 0,A
			case 0x88: bc.setH(res(bc.getH(), 1)); break;							// RES 1,B
			case 0x89: bc.setL(res(bc.getL(),   1)); break;						// RES 1,C
			case 0x8A: de.setH(res(de.getH(), 1)); break;							// RES 1,D
			case 0x8B: de.setL(res(de.getL(),   1)); break;						// RES 1,E
			case 0x8C: hl.setH(res(hl.getH(), 1)); break;							// RES 1,H
			case 0x8D: hl.setL(res(hl.getL(),   1)); break;						// RES 1,L
			case 0x8E: mem.write(hl.get(), res(mem.read(hl.get()), 1)); break;	// RES 1,(HL)
			case 0x8F: a.set(res(a.get(),           1)); break;					// RES 1,A
			case 0x90: bc.setH(res(bc.getH(), 2)); break;							// RES 2,B
			case 0x91: bc.setL(res(bc.getL(),   2)); break;						// RES 2,C
			case 0x92: de.setH(res(de.getH(), 2)); break;							// RES 2,D
			case 0x93: de.setL(res(de.getL(),   2)); break;						// RES 2,E
			case 0x94: hl.setH(res(hl.getH(), 2)); break;							// RES 2,H
			case 0x95: hl.setL(res(hl.getL(),   2)); break;						// RES 2,L
			case 0x96: mem.write(hl.get(), res(mem.read(hl.get()), 2)); break;	// RES 2,(HL)
			case 0x97: a.set(res(a.get(),           2)); break;					// RES 2,A
			case 0x98: bc.setH(res(bc.getH(), 3)); break;							// RES 3,B
			case 0x99: bc.setL(res(bc.getL(),   3)); break;						// RES 3,C
			case 0x9A: de.setH(res(de.getH(), 3)); break;							// RES 3,D
			case 0x9B: de.setL(res(de.getL(),   3)); break;						// RES 3,E
			case 0x9C: hl.setH(res(hl.getH(), 3)); break;							// RES 3,H
			case 0x9D: hl.setL(res(hl.getL(),   3)); break;						// RES 3,L
			case 0x9E: mem.write(hl.get(), res(mem.read(hl.get()), 3)); break;	// RES 3,(HL)
			case 0x9F: a.set(res(a.get(),           3)); break;					// RES 3,A
			case 0xA0: bc.setH(res(bc.getH(), 4)); break;							// RES 4,B
			case 0xA1: bc.setL(res(bc.getL(),   4)); break;						// RES 4,C
			case 0xA2: de.setH(res(de.getH(), 4)); break;							// RES 4,D
			case 0xA3: de.setL(res(de.getL(),   4)); break;						// RES 4,E
			case 0xA4: hl.setH(res(hl.getH(), 4)); break;							// RES 4,H
			case 0xA5: hl.setL(res(hl.getL(),   4)); break;						// RES 4,L
			case 0xA6: mem.write(hl.get(), res(mem.read(hl.get()), 4)); break;	// RES 4,(HL)
			case 0xA7: a.set(res(a.get(),           4)); break;					// RES 4,A
			case 0xA8: bc.setH(res(bc.getH(), 5)); break;							// RES 5,B
			case 0xA9: bc.setL(res(bc.getL(),   5)); break;						// RES 5,C
			case 0xAA: de.setH(res(de.getH(), 5)); break;							// RES 5,D
			case 0xAB: de.setL(res(de.getL(),   5)); break;						// RES 5,E
			case 0xAC: hl.setH(res(hl.getH(), 5)); break;							// RES 5,H
			case 0xAD: hl.setL(res(hl.getL(),   5)); break;						// RES 5,L
			case 0xAE: mem.write(hl.get(), res(mem.read(hl.get()), 5)); break;	// RES 5,(HL)
			case 0xAF: a.set(res(a.get(),           5)); break;					// RES 5,A
			case 0xB0: bc.setH(res(bc.getH(), 6)); break;							// RES 6,B
			case 0xB1: bc.setL(res(bc.getL(),   6)); break;						// RES 6,C
			case 0xB2: de.setH(res(de.getH(), 6)); break;							// RES 6,D
			case 0xB3: de.setL(res(de.getL(),   6)); break;						// RES 6,E
			case 0xB4: hl.setH(res(hl.getH(), 6)); break;							// RES 6,H
			case 0xB5: hl.setL(res(hl.getL(),   6)); break;						// RES 6,L
			case 0xB6: mem.write(hl.get(), res(mem.read(hl.get()), 6)); break;	// RES 6,(HL)
			case 0xB7: a.set(res(a.get(),           6)); break;					// RES 6,A
			case 0xB8: bc.setH(res(bc.getH(), 7)); break;							// RES 7,B
			case 0xB9: bc.setL(res(bc.getL(),   7)); break;						// RES 7,C
			case 0xBA: de.setH(res(de.getH(), 7)); break;							// RES 7,D
			case 0xBB: de.setL(res(de.getL(),   7)); break;						// RES 7,E
			case 0xBC: hl.setH(res(hl.getH(), 7)); break;							// RES 7,H
			case 0xBD: hl.setL(res(hl.getL(),   7)); break;						// RES 7,L
			case 0xBE: mem.write(hl.get(), res(mem.read(hl.get()), 7)); break;	// RES 7,(HL)
			case 0xBF: a.set(res(a.get(),           7)); break;					// RES 7,A
			case 0xC0: bc.setH(setBit(bc.getH(), 0)); break; 						// SET 0,B
			case 0xC1: bc.setL(setBit(bc.getL(), 0)); break;						// SET 0,C
			case 0xC2: de.setH(setBit(de.getH(), 0)); break; 						// SET 0,D
			case 0xC3: de.setL(setBit(de.getL(), 0)); break;						// SET 0,E
			case 0xC4: hl.setH(setBit(hl.getH(), 0)); break; 						// SET 0,H
			case 0xC5: hl.setL(setBit(hl.getL(), 0)); break;						// SET 0,L
			case 0xC6: mem.write(hl.get(), setBit(mem.read(hl.get()), 0)); break;	// SET 0,(HL)
			case 0xC7: a.set(setBit(a.get(), 0)); break;							// SET 0,A
			case 0xC8: bc.setH(setBit(bc.getH(), 1)); break; 						// SET 1,B
			case 0xC9: bc.setL(setBit(bc.getL(), 1)); break;						// SET 1,C
			case 0xCA: de.setH(setBit(de.getH(), 1)); break; 						// SET 1,D
			case 0xCB: de.setL(setBit(de.getL(), 1)); break;						// SET 1,E
			case 0xCC: hl.setH(setBit(hl.getH(), 1)); break; 						// SET 1,H
			case 0xCD: hl.setL(setBit(hl.getL(), 1)); break;						// SET 1,L
			case 0xCE: mem.write(hl.get(), setBit(mem.read(hl.get()), 1)); break;	// SET 1,(HL)
			case 0xCF: a.set(setBit(a.get(), 1)); break;							// SET 1,A
			case 0xD0: bc.setH(setBit(bc.getH(), 2)); break; 						// SET 2,B
			case 0xD1: bc.setL(setBit(bc.getL(), 2)); break;						// SET 2,C
			case 0xD2: de.setH(setBit(de.getH(), 2)); break; 						// SET 2,D
			case 0xD3: de.setL(setBit(de.getL(), 2)); break;						// SET 2,E
			case 0xD4: hl.setH(setBit(hl.getH(), 2)); break; 						// SET 2,H
			case 0xD5: hl.setL(setBit(hl.getL(), 2)); break;						// SET 2,L
			case 0xD6: mem.write(hl.get(), setBit(mem.read(hl.get()), 2)); break;	// SET 2,(HL)
			case 0xD7: a.set(setBit(a.get(), 2)); break;							// SET 2,A
			case 0xD8: bc.setH(setBit(bc.getH(), 3)); break; 						// SET 3,B
			case 0xD9: bc.setL(setBit(bc.getL(), 3)); break;						// SET 3,C
			case 0xDA: de.setH(setBit(de.getH(), 3)); break; 						// SET 3,D
			case 0xDB: de.setL(setBit(de.getL(), 3)); break;						// SET 3,E
			case 0xDC: hl.setH(setBit(hl.getH(), 3)); break; 						// SET 3,H
			case 0xDD: hl.setL(setBit(hl.getL(), 3)); break;						// SET 3,L
			case 0xDE: mem.write(hl.get(), setBit(mem.read(hl.get()), 3)); break;	// SET 3,(HL)
			case 0xDF: a.set(setBit(a.get(), 3)); break;							// SET 3,A
			case 0xE0: bc.setH(setBit(bc.getH(), 4)); break; 						// SET 4,B
			case 0xE1: bc.setL(setBit(bc.getL(), 4)); break;						// SET 4,C
			case 0xE2: de.setH(setBit(de.getH(), 4)); break; 						// SET 4,D
			case 0xE3: de.setL(setBit(de.getL(), 4)); break;						// SET 4,E
			case 0xE4: hl.setH(setBit(hl.getH(), 4)); break; 						// SET 4,H
			case 0xE5: hl.setL(setBit(hl.getL(), 4)); break;						// SET 4,L
			case 0xE6: mem.write(hl.get(), setBit(mem.read(hl.get()), 4)); break;	// SET 4,(HL)
			case 0xE7: a.set(setBit(a.get(), 4)); break;							// SET 4,A
			case 0xE8: bc.setH(setBit(bc.getH(), 5)); break; 						// SET 5,B
			case 0xE9: bc.setL(setBit(bc.getL(), 5)); break;						// SET 5,C
			case 0xEA: de.setH(setBit(de.getH(), 5)); break; 						// SET 5,D
			case 0xEB: de.setL(setBit(de.getL(), 5)); break;						// SET 5,E
			case 0xEC: hl.setH(setBit(hl.getH(), 5)); break; 						// SET 5,H
			case 0xED: hl.setL(setBit(hl.getL(), 5)); break;						// SET 5,L
			case 0xEE: mem.write(hl.get(), setBit(mem.read(hl.get()), 5)); break;	// SET 5,(HL)
			case 0xEF: a.set(setBit(a.get(), 5)); break;							// SET 5,A
			case 0xF0: bc.setH(setBit(bc.getH(), 6)); break; 						// SET 6,B
			case 0xF1: bc.setL(setBit(bc.getL(), 6)); break;						// SET 6,C
			case 0xF2: de.setH(setBit(de.getH(), 6)); break; 						// SET 6,D
			case 0xF3: de.setL(setBit(de.getL(), 6)); break;						// SET 6,E
			case 0xF4: hl.setH(setBit(hl.getH(), 6)); break; 						// SET 6,H
			case 0xF5: hl.setL(setBit(hl.getL(), 6)); break;						// SET 6,L
			case 0xF6: mem.write(hl.get(), setBit(mem.read(hl.get()), 6)); break;	// SET 6,(HL)
			case 0xF7: a.set(setBit(a.get(), 6)); break;							// SET 6,A
			case 0xF8: bc.setH(setBit(bc.getH(), 7)); break; 						// SET 7,B
			case 0xF9: bc.setL(setBit(bc.getL(), 7)); break;						// SET 7,C
			case 0xFA: de.setH(setBit(de.getH(), 7)); break; 						// SET 7,D
			case 0xFB: de.setL(setBit(de.getL(), 7)); break;						// SET 7,E
			case 0xFC: hl.setH(setBit(hl.getH(), 7)); break; 						// SET 7,H
			case 0xFD: hl.setL(setBit(hl.getL(), 7)); break;						// SET 7,L
			case 0xFE: mem.write(hl.get(), setBit(mem.read(hl.get()), 7)); break;	// SET 7,(HL)
			case 0xFF: a.set(setBit(a.get(), 7)); break;							// SET 7,A

			// Unimplented CB Opcode
			default:
				System.out.println("Unimplemented CB Opcode: "+Integer.toHexString(opcode));
				break;
		}
		pc++; 			// All CB opcodes are one byte (CB n).
	}


	/**
	 *	CB RLC - Rotate Left Carry
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int rlc(int value)
	{
		int temp = value;
		value = (value << 1)&0xff;
		if ((temp & 0x80) == 0x80)
		{
			f.carryOn();
			value |= 0x01;
		}
		else
		{
			f.carryOff();
		}

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}


	/**
	 *	CB RRC - Rotate Right Carry
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int rrc(int value)
	{
		int temp = value;
		value >>= 1;
		if ((temp & 0x01) == 0x01)
		{
			f.carryOn();
			value |= 0x80;
		}
		else
		{
			f.carryOff();
		}

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.setParity(value); f.negativeOff(); f.halfCarryOff();
		return value;
	}


	/**
	 *	CB RL - Rotate Left
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int rl(int value)
	{
		int temp = value;
		// Shift Left One Bit Position
		value = (value << 1)&0xff;

		// Move contents of carry flag to bit 0
		if (f.carry()) value |= 0x01;

		// Move original contents to carry flag
		if ((temp & 0x80) == 0x80) f.carryOn();
		else f.carryOff();
		
		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}


	/**
	 *	CB RR - Rotate Right
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int rr(int value)
	{
		int temp = value;

		// Shift Right One Bit Position
		value >>= 1;

		// Move contents of carry flag to bit 7
		if (f.carry()) value |= 0x80;

		// Move original contents to carry flag
		if ((temp & 0x01) != 0) f.carryOn();
		else f.carryOff();
		
		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}


	/**
	 *	CB SLA - Shift Left Arithmetic
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int sla(int value)
	{
		if ((value & 0x80) == 0x80) f.carryOn();
		else f.carryOff();

		value = (value<<1)&0xff;
		value &= ~0x01;

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}


	/**
	 *	CB SRA - Shift Right Arithmetic
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int sra(int value)
	{
		if ((value & 0x01) == 0x01) f.carryOn();
		else f.carryOff();

		int temp = (value & 0x80);
		value = (value >> 1) |temp;

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}
	
	/**
	 *	CB SLL - Logical Left Shift
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int sll(int value)
	{
		if ((value & 0x80) == 0x80) f.carryOn();
		else f.carryOff();

		value = (value<<1)&0xff;
		value |= 0x01;

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) != 0) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}

	/**
	 *	CB SRL - Logical Shift Right
	 *
	 *  @param value		Value to adjust
	 *
	 *	@return				Adjusted value
     */

	private final int srl(int value)
	{
		if ((value & 0x01) == 0x01) f.carryOn();
		else f.carryOff();

		value >>= 1;
		value &= ~0x80;

		if (value == 0) f.zeroOn();
		else f.zeroOff();

		if ((value & 0x80) == 0x80) f.signOn();
		else f.signOff();

		f.negativeOff(); f.setParity(value); f.halfCarryOff();
		return value;
	}


	/**
	 *	CB SET - Set Bit On
	 *
	 *  @param value		Value to adjust
	 *	@param b			Bit to turn on
	 *
	 *	@return				Adjusted value
     */

	private final int setBit(int value, int b)
	{
		b = (0x01<<b);	// Convert 0-7 to real bit
		value |= b;
		return value;
	}


	/**
	 *	CB RES - Reset Bit
	 *
	 *  @param value		Value to adjust
	 *	@param b			Bit to turn off
	 *
	 *	@return				Adjusted value
     */

	private final int res(int value, int b)
	{
		b = (0x01<<b);	// Convert 0-7 to real bit
		value &= ~b;
		return value;
	}


	/**
	 *	CB BIT - Test Bit
	 *
	 *  @param value		Value to test
	 *	@param b			Bit of value to test
     */

	private final void bit(int value, int b)
	{
		b = (0x01<<b);	// Convert 0-7 to real bit
		if ((value & b) != 0) // BIT IS SET
		{
			f.zeroOff();
			f.parityOff();
			
			// Undocumented Flags
			switch (b)
			{
				case 0x08: f.bit3On(); break;
				case 0x20: f.bit5On(); break;
				case 0x80: f.signOn(); break;
			}
		}
		else
		{
			f.zeroOn();
			f.parityOn();
		}

		f.negativeOff(); f.halfCarryOn();
	}


	/**
	 *	Execute DD/FD Prefixed Index Opcode
	 *
	 *	@param opcode		Opcode hex value
	 *  @param index		Index Register to use
     */

	private final void doIndexOp(int opcode, Registers index)
	{
		int temp;
		tstates += opddstates[opcode];
		r.inc();

		switch(opcode)
		{
			case 0x09: index.add(bc.get()); pc++; break; 						// ADD IX,BC
			case 0x19: index.add(de.get()); pc++; break; 						// ADD IX,DE
			case 0x21: index.set(mem.readWord(pc+1)); pc+=3; break; 			// LD IX,nn
			case 0x22: mem.write(mem.readWord(pc+1), index.getL());			// LD (nn),IX
					    mem.write((mem.readWord(pc+1)+1), index.getH());
					   pc+=3; break;
			case 0x23: index.inc(); pc++; break; 								// INC IX
			case 0x24: index.incH(); pc++; break; 							// INC IXh *
			case 0x25: index.decH(); pc++; break; 							// DEC IXh *
			case 0x26: index.setH(mem.read(++pc)); pc++; break;				// LD IXh,n *
			case 0x29: index.add(index.get()); pc++; break; 					// ADD IX,IX
			case 0x2A: index.setL(mem.read(mem.readWord(pc+1)));				// LD IX,(nn)
					   index.setH(mem.read(mem.readWord(pc+1)+1));
					   pc+=3; break;
			case 0x2B: index.dec(); pc++; break; 								// DEC IX
			case 0x2C: index.incL(); pc++; break; 							// INC IXl *
			case 0x2D: index.decL(); pc++; break; 							// DEC IXl *
			case 0x2E: index.setL(mem.read(++pc)); pc++; break; 				// LD IXl,n
			case 0x34: inchl(index.get()+d()); pc++; break;					// INC (IX+d)
			case 0x35: dechl(index.get()+d()); pc++; break;					// DEC (IX+d)
			case 0x36: mem.write(index.get()+d(), mem.read(++pc)); pc++; break;// LD (IX+d), n
			case 0x39: index.add(sp); pc++; break; 							// ADD IX,SP
			case 0x44: bc.setH(index.getH());	pc++; break;					// LD B,IXh *
			case 0x45: bc.setH(index.getL());	pc++; break;					// LD B,IXl *
			case 0x46: bc.setH(mem.read(index.get()+d())); 	pc++; break; 	// LD B,(IX+d)
			case 0x4C: bc.setL(index.getH());	pc++; break;					// LD C,IXh *
			case 0x4D: bc.setL(index.getL());	pc++; break;					// LD C,IXl *
			case 0x4E: bc.setL(mem.read(index.get()+d())); 	pc++; break; 	// LD C,(IX+d)
			case 0x54: de.setH(index.getH());	pc++; break;					// LD D,IXh *
			case 0x55: de.setH(index.getL());	pc++; break;					// LD D,IXl *
			case 0x56: de.setH(mem.read(index.get()+d())); 	pc++; break; 	// LD D,(IX+d)
			case 0x5C: de.setL(index.getH());	pc++; break;					// LD E,IXh *
			case 0x5D: de.setL(index.getL());	pc++; break;					// LD E,IXl *
			case 0x5E: de.setL(mem.read(index.get()+d())); 	pc++; break; 	// LD E,(IX+d)
			case 0x60: index.setH(bc.getH()); pc++; break;					// LD IXh,B *
			case 0x61: index.setH(bc.getL()); pc++; break;					// LD IXh,C *
			case 0x62: index.setH(de.getH()); pc++; break;					// LD IXh,D *
			case 0x63: index.setH(de.getL()); pc++; break;					// LD IXh,E *
			case 0x64: pc++; break;											// LD IXH,IXH*
			case 0x65: index.setH(index.getL()); pc++; break;					// LD IXH,IXL *
			case 0x66: hl.setH(mem.read(index.get()+d())); 	pc++; break; 	// LD H,(IX+d)
			case 0x67: index.setH(a.get()); pc++; break;						// LD IXh,A *
			case 0x68: index.setL(bc.getH()); pc++; break;					// LD IXL,B *
			case 0x69: index.setL(bc.getL()); pc++; break;					// LD IXl,C *
			case 0x6A: index.setL(de.getH()); pc++; break;					// LD IXL,D *
			case 0x6B: index.setL(de.getL()); pc++; break;					// LD IXl,E *
			case 0x6C: index.setL(index.getH());	pc++; break;				// LD IXl,IXh *
			case 0x6D: 									pc++; break;		// LD IXl,IXl *
			case 0x6E: hl.setL(mem.read(index.get()+d())); 	pc++; break; 	// LD L,(IX+d)
			case 0x6F: index.setL(a.get()); pc++; break; 						// LD IXl,A *
			case 0x70: mem.write(index.get()+d(), bc.getH()); pc++; break; 	// LD (IX+d),B
			case 0x71: mem.write(index.get()+d(), bc.getL()); pc++; break; 	// LD (IX+d),C
			case 0x72: mem.write(index.get()+d(), de.getH()); pc++; break; 	// LD (IX+d),D
			case 0x73: mem.write(index.get()+d(), de.getL()); pc++; break; 	// LD (IX+d),E
			case 0x74: mem.write(index.get()+d(), hl.getH()); pc++; break; 	// LD (IX+d),H
			case 0x75: mem.write(index.get()+d(), hl.getL()); pc++; break; 	// LD (IX+d),L
			case 0x77: mem.write(index.get()+d(), a.get()); pc++; break; 		// LD (IX+d),A
			case 0x7C: a.set(index.getH()); pc++; break; 						// LD A,IXh *
			case 0x7D: a.set(index.getL()); pc++; break; 						// LD A,IXl *
			case 0x7E: a.set(mem.read(index.get()+d())); pc++; break; 		// LD A,(IX+d)
			case 0x84: a.add(index.getH()); pc++; break;						// ADD A,IXh *
			case 0x85: a.add(index.getL()); pc++; break;						// ADD A,IXl *
			case 0x86: a.add(mem.read(index.get()+d())); pc++; break; 		// ADD A,(IX+d)
			case 0x8C: a.adc(index.getH()); pc++; break;						// ADC A,IXH *
			case 0x8D: a.adc(index.getL()); pc++; break;						// ADC A,IXL *
			case 0x8E: a.adc(mem.read(index.get()+d())); pc++; break; 		// ADC A,(IX+d)
			case 0x94: a.sub(index.getH()); pc++; break;						// SUB IXh *
			case 0x95: a.sub(index.getL()); pc++; break;						// SUB IXl *
			case 0x96: a.sub(mem.read(index.get()+d())); pc++; break; 		// SUB A,(IX+d)
			case 0x9C: a.sbc(index.getH()); pc++; break;						// SBC A,IXH *
			case 0x9D: a.sbc(index.getL()); pc++; break;						// SBC A,IXL *
			case 0x9E: a.sbc(mem.read(index.get()+d())); pc++; break; 		// SBC A,(IX+d)
			case 0xA4: a.and(index.getH()); pc++; break;						// AND IXh *
			case 0xA5: a.and(index.getL()); pc++; break;						// AND IXl *
			case 0xA6: a.and(mem.read(index.get()+d())); pc++; break; 		// AND A,(IX+d)
			case 0xAC: a.xor(index.getH()); pc++; break; 						// XOR A IXH*
			case 0xAD: a.xor(index.getL()); pc++; break; 						// XOR A IXL*
			case 0xAE: a.xor(mem.read(index.get()+d())); pc++; break; 		// XOR A,(IX+d)
			case 0xB4: a.or(index.getH()); pc++; break; 						// OR A IXH*
			case 0xB5: a.or(index.getL()); pc++; break; 						// OR A IXL*
			case 0xB6: a.or(mem.read(index.get()+d())); pc++; break; 			// OR A,(IX+d)
			case 0xBC: a.cp(index.getH()); pc++; break;						// CP IXh *
			case 0xBD: a.cp(index.getL()); pc++; break;						// CP IXl *
			case 0xBE: a.cp(mem.read(index.get()+d())); pc++; break; 			// CP (IX+d)
			case 0xCB: doIndexCB(index); break;								// CB Opcode
			case 0xE1: index.set(pop()); pc++; break; 						// POP IX
			case 0xE3: temp = index.get(); index.set(mem.readWord(sp));		// EX SP,(IX)
					   mem.write(sp, temp&0xff); mem.write(sp+1, temp>>8);
					   pc++; break;
			case 0xE5: push(index.get());				pc++; break;			// PUSH IX
			case 0xE9: pc = index.get();					  break; 			// JP (IX)
			case 0xF9: sp = index.get();				pc++; break; 			// LD SP,IX

			// Unimplented DD/FD Opcode
			default:
				System.out.println("Unimplemented DD or FD Opcode: "+Integer.toHexString(opcode));
				pc++;
				break;
		} // end of switch
	}

	/**
	 *	Get Offset Value from memory for index operations
	 *
	 *  @return				Offset
     */

	private final int d()
	{
		return mem.readSigned(++pc);
	}


	/**
	 *	Execute DDCB/FDCB Prefixed Opcode
	 *
	 *	@param index		Index Register To Use
     */

	private final void doIndexCB(Registers index)
	{
		int location = (index.get() + d())&0xFFFF;
		int opcode = mem.read(++pc);
		tstates += opindexcbstates[opcode];

		switch (opcode)
		{
			case 0x06: mem.write(location, rlc(mem.read(location))); break;	// RLC (IX)
			case 0x0E: mem.write(location, rrc(mem.read(location)));	break;	// RRC (IX)
			case 0x16: mem.write(location, rl(mem.read(location))); break;	// RL (IX)
			case 0x1E: mem.write(location, rr(mem.read(location))); break;	// RR (IX)
			case 0x26: mem.write(location, sla(mem.read(location))); break;	// SLA (IX)
			case 0x2E: mem.write(location, sra(mem.read(location))); break;	// SRA (IX)
			case 0x36: mem.write(location, sll(mem.read(location))); break;	// SLL (IX) *
			case 0x3E: mem.write(location, srl(mem.read(location))); break;	// SRL (IX)
			case 0x46: bit(mem.read(location), 0); break;						// BIT 0,(IX)
			case 0x4E: bit(mem.read(location), 1); break;						// BIT 1,(IX)
			case 0x56: bit(mem.read(location), 2); break;						// BIT 2,(IX)
			case 0x5E: bit(mem.read(location), 3); break;						// BIT 3,(IX)
			case 0x66: bit(mem.read(location), 4); break;						// BIT 4,(IX)
			case 0x6E: bit(mem.read(location), 5); break;						// BIT 5,(IX)
			case 0x76: bit(mem.read(location), 6); break;						// BIT 6,(IX)
			case 0x7E: bit(mem.read(location), 7); break;						// BIT 7,(IX)
			case 0x86: mem.write(location, res(mem.read(location), 0)); break; // RES 0,(IX)
			case 0x8E: mem.write(location, res(mem.read(location), 1)); break; // RES 1,(IX)
			case 0x96: mem.write(location, res(mem.read(location), 2)); break; // RES 2,(IX)
			case 0x9E: mem.write(location, res(mem.read(location), 3)); break; // RES 3,(IX)
			case 0xA6: mem.write(location, res(mem.read(location), 4)); break; // RES 4,(IX)
			case 0xAE: mem.write(location, res(mem.read(location), 5)); break; // RES 5,(IX)
			case 0xB6: mem.write(location, res(mem.read(location), 6)); break; // RES 6,(IX)
			case 0xBE: mem.write(location, res(mem.read(location), 7)); break; // RES 7,(IX)
			case 0xC6: mem.write(location, setBit(mem.read(location), 0)); break;// SET 0,(IX)
			case 0xCE: mem.write(location, setBit(mem.read(location), 1)); break;// SET 1,(IX)
			case 0xD6: mem.write(location, setBit(mem.read(location), 2)); break;// SET 2,(IX)
			case 0xDE: mem.write(location, setBit(mem.read(location), 3)); break;// SET 3,(IX)
			case 0xE6: mem.write(location, setBit(mem.read(location), 4)); break;// SET 4,(IX)
			case 0xEE: mem.write(location, setBit(mem.read(location), 5)); break;// SET 5,(IX)
			case 0xF6: mem.write(location, setBit(mem.read(location), 6)); break;// SET 6,(IX)
			case 0xFE: mem.write(location, setBit(mem.read(location), 7)); break;// SET 7,(IX)

			default:
				System.out.println("Unimplemented DDCB or FDCB Opcode: "+Integer.toHexString(opcode&0xff));
				break;

		} // end of switch
		pc++;
	}


	/**
	 *	Execute ED Prefixed Opcode
	 *
	 *	@param opcode		Opcode hex value
     */

	public final void doED(int opcode)
	{
		int temp;
		int temp2;
		boolean carry; // to preserve carry flag

		tstates += opedstates[opcode];
		r.inc();

		switch(opcode)
		{

			//  -- ED40 IN B,(C) -------------------------
			case 0x40:
				temp = port.in(bc.getL());
				bc.setH(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) != 0) f.signOn();
				else f.signOff();
				f.setParity(temp); f.negativeOff(); f.halfCarryOff();
				pc++;
				break;

			//  -- ED41 OUT (C),B -------------------------
			case 0x41:
				port.out(bc.getL(), bc.getH());
				pc++;
				break;

			// --  ED42 SBC HL, BC ------------------------ (C)
			case 0x42:
				hl.sbc(bc.get());
				pc++;
				break;

			//  -- ED43 LD (nn),BC ------------------------
			case 0x43:
				temp = mem.readWord(pc+1);
				mem.write(temp, bc.getL());
				mem.write(temp+1, bc.getH());
				pc+=3;
				break;

			//  -- ED44 NEG -------------------------------
			case 0x44:
			case 0x4C:
			case 0x54:
			case 0x5C:
			case 0x64:
			case 0x6C:
			case 0x74:
			case 0x7C:
				// A <- 0-A
				temp = a.get();
				a.set(0);
				a.sub(temp);
				pc++;
				break;

			//  -- ED45 RETN / RETI ------------------------------
			case 0x45:
			case 0x4D:
			case 0x55:
			case 0x5D:
			case 0x65:
			case 0x6D:
			case 0x75:
			case 0x7D:
				pc = pop();
				iff1 = iff2;
				break;

			//  -- ED46 IM 0-------------------------------
			case 0x46:
			case 0x4E:
			case 0x66:
			case 0x6E:
				im = 0;
				pc++;
				break;

			//  -- ED47 LD I, A ---------------------------
			case 0x47:
				i = a.get();
				pc++;
				break;

			//  -- ED48 IN C,(C) -------------------------
			case 0x48:
				temp = port.in(bc.getL());
				bc.setL(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) != 0) f.signOn();
				else f.signOff();
				f.setParity(temp); f.negativeOff(); f.halfCarryOff();
				pc++;
				break;

			//  -- ED49 OUT (C),C -------------------------
			case 0x49:
				port.out(bc.getL(), bc.getL());
				pc++;
				break;

			//  -- ED4A ADC HL, BC ------------------------
			case 0x4A:
				hl.adc(bc.get());
				pc++;
				break;

			//  -- ED4B LD BC, (nn) -----------------------
			case 0x4B:
				temp = mem.readWord(pc+1);
				bc.setL(mem.read(temp));
				bc.setH(mem.read(temp+1));
				pc+=3;
				break;

			//  -- ED4F LD R, A ---------------------------
			case 0x4F:
				r.set(a.get());
				pc++;
				break;

			//  -- ED50 IN D,(C) -------------------------
			case 0x50:
				temp = port.in(bc.getL());
				de.setH(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) != 0) f.signOn();
				else f.signOff();
				f.setParity(temp); f.negativeOff(); f.halfCarryOff();;
				pc++;
				break;

			//  -- ED51 OUT (C),D -------------------------
			case 0x51:
				port.out(bc.getL(), de.getH());
				pc++;
				break;

			// --  ED52 SBC HL, DE ------------------------
			case 0x52:
				hl.sbc(de.get());
				pc++;
				break;

			//  -- ED53 LD (nn),DE ------------------------
			case 0x53:
				temp = mem.readWord(pc+1);
				mem.write(temp, de.getL());		//SPl
				mem.write(temp+1, de.getH()); 	//SPh
				pc+=3;
				break;

			//  -- ED56 IM 1-------------------------------
			case 0x56:
			case 0x76:
				im = 1;
				pc++;
				break;

			//  -- ED57 LD A, I ---------------------------
			case 0x57:
				a.set(i);
				if (iff2) f.parityOn();
				else f.parityOff();
				if (i == 0) f.zeroOn();
				else f.zeroOff();
				if ((i & 0x80) != 0) f.signOn();
				else f.signOff();
				f.halfCarryOff(); f.negativeOff();
				pc++;
				break;

			//  -- ED58 IN E,(C) -------------------------
			case 0x58:
				temp = port.in(bc.getL());
				de.setL(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) != 0) f.signOn();
				else f.signOff();
				f.setParity(temp); f.negativeOff(); f.halfCarryOff();
				pc++;
				break;

			//  -- ED59 OUT (C),E -------------------------
			case 0x59:
				port.out(bc.getL(), de.getL());
				pc++;
				break;

			//  -- ED5A ADC HL, DE ------------------------
			case 0x5A:
				hl.adc(de.get());
				pc++;
				break;

			//  -- ED5B LD DE, (nn) -----------------------
			case 0x5B:
				temp = mem.readWord(pc+1);
				de.setL(mem.read(temp));
				de.setH(mem.read(temp+1));
				pc+=3;
				break;

			// -- ED5F LD A,R -----------------------------
			case 0x5F:
				temp = r.get();
				a.set(temp);
				if (iff2) f.parityOn();
				else f.parityOff();
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) != 0) f.signOn();
				else f.signOff();
				f.halfCarryOff(); f.negativeOff();
				pc++;
				break;

			//  -- ED60 IN H,(C) -------------------------
			case 0x60:
				temp = port.in(bc.getL());
				hl.setH(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) == 0x80) f.signOn();
				else f.signOff();
				f.setParity(temp); f.negativeOff(); f.halfCarryOff();
				pc++;
				break;

			//  -- ED61 OUT (C),H -------------------------
			case 0x61:
				port.out(bc.getL(), hl.getH());
				pc++;
				break;

			// --  ED62 SBC HL, HL ------------------------
			case 0x62:
				hl.sbc(hl.get());
				pc++;
				break;

			//  -- ED63 LD (nn),HL ------------------------
			case 0x63:
				mem.write(mem.readWord(pc+1), hl.getL());		//SPl
				mem.write((mem.readWord(pc+1)+1), hl.getH()); 	//SPh
				pc+=3;
				break;

			//  -- ED67 RRD -------------------------------
			case 0x67:
				temp = a.get();
				temp2 = mem.read(hl.get());
				// move high 4 of hl to low 4 of hl
				// move low 4 of a to high 4 of hl
				mem.write(hl.get(), (temp2>>4) | ((temp&0x0f)<<4));
				// move 4 lowest bits of hl to low 4 of a
				temp = (temp&0xF0) | (temp2&0x0F);
				a.set(temp);

				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) == 0) f.signOff();
				else f.signOn();

				f.setParity(temp);

				f.negativeOff(); f.halfCarryOff();
				pc++;
				break;

			//  -- ED68 IN L,(C) --------------------------
			case 0x68:
				temp = port.in(bc.getL());
				hl.setL(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) == 0x80) f.signOn();
				else f.signOff();
				f.setParity(temp); f.negativeOff(); f.halfCarryOff();
				pc++;
				break;

			//  -- ED69 OUT (C),L -------------------------
			case 0x69:
				port.out(bc.getL(), hl.getL());
				pc++;
				break;

			//  -- ED6A ADC HL, HL ------------------------
			case 0x6A:
				hl.adc(hl.get());
				pc++;
				break;

			//  -- ED6B LD HL,(nn) -----------------------
			case 0x6B:
				temp = mem.readWord(pc+1);
				hl.setL(mem.read(temp));
				hl.setH(mem.read(temp+1));
				pc+=3;
				break;

			//  -- ED6F RLD -------------------------------
			case 0x6F:
				temp = a.get();
				temp2 = mem.read(hl.get());
				// move low 4 of hl to high 4 of hl
				// move low 4 of a to low 4 of hl
				mem.write(hl.get(), (temp2&0x0F)<<4 | (temp&0x0F));
				// move high 4 of hl to low 4 of a
				temp = (temp&0xF0) | (temp2>>4);
				a.set(temp);

				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) == 0) f.signOff();
				else f.signOn();

				f.negativeOff(); f.halfCarryOff();
				f.setParity(temp);
				pc++;
				break;

			//  *- ED71 OUT (C),0 -------------------------
			case 0x71:
				port.out(bc.getL(), 0);
				pc++;
				break;

			// --  ED72 SBC HL, SP ------------------------
			case 0x72:
				hl.sbc(sp);
				pc++;
				break;

			//  -- ED73 LD (nn),SP ------------------------
			case 0x73:
				mem.write(mem.readWord(pc+1), sp&0xff);		//SPl
				mem.write((mem.readWord(pc+1)+1), sp>>8); 	//SPh
				pc+=3;
				break;

			//  -- ED78 IN A,(C) -------------------------
			// Output reg A to Port stored in reg c
			case 0x78:
				temp = port.in(bc.getL());
				a.set(temp);
				if (temp == 0) f.zeroOn();
				else f.zeroOff();
				if ((temp & 0x80) == 0x80) f.signOn();
				else f.signOff();
				f.setParity(temp);
				f.negativeOff();
				f.halfCarryOff();;
				pc++;
				break;

			//  -- ED79 OUT (C),A -------------------------
			// Output reg A to Port stored in reg c
			case 0x79:
				port.out(bc.getL(), a.get());
				pc++;
				break;

			// --  ED7A ADC HL, SP ------------------------
			case 0x7A:
				hl.adc(sp);
				pc++;
				break;

			//  -- ED7B LD SP, (nn) -----------------------
			case 0x7B:
				temp = mem.read(mem.readWord(pc+1)); // SPl
				temp2 = mem.read(mem.readWord(pc+1)+1); // SPh

				sp = (temp&0xff) | (temp2<<8);
				pc+=3;
				break;

			//  -- EDA0 LDI ---------------------------------- (c)
			case 0xA0:
				// (DE) <- (HL)
				mem.write(de.get(), mem.read(hl.get()));
				de.inc(); hl.inc(); bc.dec();

				if (bc.get() == 0) f.parityOff();
				else f.parityOn();
				f.negativeOff(); f.halfCarryOff();;
				pc++;
				break;

			//  -- EDA1 CPI ------------------------------
			case 0xA1:
				carry = f.carry();
				a.cp(mem.read(hl.get()));
				hl.inc();
				bc.dec();

				if(bc.get() == 0) f.parityOff();
				else f.parityOn();

				f.negativeOn();

				if (carry) f.carryOn();
				else f.carryOff();

				pc++;
				break;

			//  -- EDA2 INI -------------------------------
			case 0xA2:
				temp = port.in(bc.getL());
				mem.write(hl.get(), temp);
				bc.decH();
				hl.inc();
				if (temp + ((bc.getH()+1)&0xFF) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();
				}
				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				pc++;
				break;

			//  -- EDA3 OUTI ------------------------------
			// see p14 of undocumented z80 for additional flag info
			case 0xA3:
				temp = mem.read(hl.get());
				// (C) <- (HL)
				port.out(bc.getL(), temp);
				// HL <- HL + 1
				hl.inc();
				// B <- B -1
				bc.decH(); // Flags in OUTI adjusted in same way as dec b anyway.
				if ((hl.getL() + temp) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();;
				}
				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				pc++;
				break;

			//  -- EDA8 LDD ----------------------------------
			case 0xA8:
				// (DE) <- (HL)
				mem.write(de.get(), mem.read(hl.get()));
				de.dec(); hl.dec(); bc.dec();

				if (bc.get() == 0) f.parityOff();
				else f.parityOn();
				f.negativeOff(); f.halfCarryOff();
				pc++;
				break;


			//  -- EDA9 CPD ------------------------------
			case 0xA9:
				carry = f.carry();
				a.cp(mem.read(hl.get()));
				hl.dec();
				bc.dec();

				if(bc.get() == 0) f.parityOff();
				else f.parityOn();

				f.negativeOn();

				if (carry) f.carryOn();
				else f.carryOff();

				pc++;
				break;


			//  -- EDAA IND -------------------------------
			case 0xAA:
				temp = port.in(bc.getL());
				mem.write(hl.get(), temp);
				bc.decH();
				hl.dec();
				if (temp + ((bc.getH()-1)&0xFF) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();
				}
				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				pc++;
				break;

			//  -- EDAB OUTD ------------------------------
			// see p14 of undocumented z80 for additional flag info
			case 0xAB:
				temp = mem.read(hl.get());
				// (C) <- (HL)
				port.out(bc.getL(), temp);
				// HL <- HL - 1
				hl.dec();
				// B <- B -1
				bc.decH(); // Flags in OUTI adjusted in same way as dec b anyway.

				if ((hl.getL() + temp) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();;
				}
				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				pc++;
				break;

			//  -- EDB0 LDIR ------------------------------
			case 0xB0:
				// (DE) <- (HL) mem.write(location, value)
				mem.write(de.get(), mem.read(hl.get()));
				// DE <- DE + 1
				de.inc();
				// HL <- HL + 1
				hl.inc();
				// BC <- BC - 1
				bc.dec();

				if (bc.get() != 0)
				{
					f.parityOn();
					tstates+=5;
					pc--;
				}
				else
				{
					f.parityOff();
					pc++;
				}

				f.negativeOff(); f.halfCarryOff();;
				break;

			//  -- EDB1 CPIR ------------------------------
			case 0xB1:
				carry = f.carry();
				a.cp(mem.read(hl.get())); // sets zero flag for us
				hl.inc();
				bc.dec();

				if (bc.get() == 0) f.parityOff();
				else f.parityOn();

				if ((f.parity()) && (!f.zero()))
				{
					tstates +=5;
					pc--;
				}
				else
				{
					pc++;
				}

				if (carry) f.carryOn();
				else f.carryOff();

				f.negativeOn(); // Sign set by the cp instruction
				break;

			//  -- EDB2 INIR ------------------------------
			case 0xB2:
				temp = port.in(bc.getL());
				mem.write(hl.get(), temp);
				bc.decH();
				hl.inc();
				if (bc.getH() != 0)
				{
					tstates+=5;
					pc--;
				}
				else
				{
					pc++;
				}
				if (temp + ((bc.getH()+1)&0xFF) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();
				}
				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				break;

			//  -- EDB3 OTIR ------------------------------
			case 0xB3:
				temp = mem.read(hl.get());
				// (C) <- (HL)
				port.out(bc.getL(), temp);
				// B <- B -1
				bc.decH();
				// HL <- HL + 1
				hl.inc();

				if (bc.getH() != 0)
				{
					tstates+=5;
					pc--;
				}
				else
				{
					pc++;
				}
				if ((hl.getL() + temp) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();;
				}

				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				break;

			// -- EDB8 LDDR ---------------------------------
			case 0xB8:
				mem.write(de.get(), mem.read(hl.get()));
				de.dec(); hl.dec(); bc.dec();

				if (bc.get() != 0)
				{
					f.parityOn();
					tstates+=5;
					pc--;
				}
				else
				{
					f.parityOff();
					pc++;
				}

				f.negativeOff(); f.halfCarryOff();;
				break;

			// -- EDB9 CPDR ------------------------------------
			case 0xB9:
				carry = f.carry();
				a.cp(mem.read(hl.get())); // sets zero flag for us
				hl.dec();
				bc.dec();

				if (bc.get() == 0) f.parityOff();
				else f.parityOn();

				if ((f.parity()) && (!f.zero()))
				{
					tstates +=5;
					pc--;
				}
				else
				{
					pc++;
				}
				if (carry) f.carryOn();
				else f.carryOff();

				f.negativeOn();
				break;

			//  -- EDBA INDR ------------------------------
			case 0xBA:
				temp = port.in(bc.getL());
				mem.write(hl.get(), temp);
				bc.decH();
				hl.dec();
				if (bc.getH() != 0)
				{
					tstates+=5;
					pc--;
				}
				else
				{
					pc++;
				}
				
				if (temp + ((bc.getH()-1)&0xFF) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();
				}

				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				break;

			//  -- EDBB OTDR ------------------------------
			case 0xBB:
				temp = mem.read(hl.get());
				// (C) <- (HL)
				port.out(bc.getL(), temp);
				// B <- B -1
				bc.decH();
				// HL <- HL + 1
				hl.dec();

				if (bc.getH() != 0)
				{
					tstates+=5;
					pc--;
				}
				else
				{
					pc++;
				}
				if ((hl.getL() + temp) > 255)
				{
					f.carryOn(); f.halfCarryOn();
				}
				else
				{
					f.carryOff(); f.halfCarryOff();
				}

				if ((temp & 0x80) == 0x80) f.negativeOn();
				else f.negativeOff();
				break;

				// -- Unimplented ED Opcode --------------------
				default:
					System.out.println("Unimplemented ED Opcode: "+Integer.toHexString(opcode));
					pc++;
					break;
		} // end of switch
	} // end of ed ops


	/**
	 * Pre-calculate DAA Table
	 */

	private void generateDAATable()
	{
		daa = new int[0x800];

		final int NEGATIVE = 0x100;
		final int CARRY = 0x200;
		final int HALFCARRY = 0x400;

		for (int x = 0; x <= 0x7FF; x++)
		{
			int a = x & 0xFF;
			int highNibble = (a >> 4);
			int lowNibble  = (a & 0x0F);
			boolean carry = false;

			if ((x & NEGATIVE) == 0) // ADD, ADC, INC
			{
				if ((x & CARRY) == 0)
				{
					if ((x & HALFCARRY) == 0)
					{
						if ((highNibble <= 0x08) && (lowNibble <= 0x0F) && (lowNibble >= 0x0A)) // Line 2
							a+=0x06;
						else if ((highNibble <= 0x09) && (lowNibble <= 0x09)); // Line 1 of Page 236
						else if ((highNibble >= 0x0A) && (lowNibble <= 0x09))  // Line 4
						{
							a+=0x60;
							carry = true;
						}
						else if ((highNibble >= 0x09) && (lowNibble >= 0x0A)) // Line 5
						{
							a+=0x66;
							carry = true;
						}
					}
					else // Half Carry is 1, Carry is 0, Negative is 0
					{
						if ((highNibble <= 0x09) && (lowNibble <= 0x03)) // Line 3
							a+=0x06;
						else if ((highNibble >= 0x0A) && (lowNibble <= 0x03)) // Line 6
						{
							a+=0x66;
							carry = true;
						}
					}

				}
				else // Flag Carry is 1
				{
					if ((x & HALFCARRY) == 0) // C = 1, HC = 0, N = 0
					{
						if ((highNibble <= 0x02) && (lowNibble <= 0x09)) // Line 7
						{
							a+=0x60;
							carry = true;
						}
						else if ((highNibble <= 0x02) && (lowNibble >= 0x0A)) // Line 8
						{
							a+=0x66;
							carry = true;
						}

					}
					else // Half Carry is 1
					{
						if ((highNibble < 0x04) && (lowNibble < 0x04)) // Line 9
						{
							a+=0x66;
							carry = true;
						}
					}
				}
			}
			// Negative Flag is true SUB, SBC, DEC, NEG --------------------------------------
			else
			{
				if ((x & CARRY) == 0)
				{
					if ((x & HALFCARRY) == 0)
					{
						if ((highNibble < 0x0A) && (lowNibble < 0x0A) );  // Line 10
					}
					else // Flag HalfCarry is 1, Carry = 0, Neg = 1
					{
						if ((highNibble < 0x09) && (lowNibble > 0x05)) // Line 11
							a +=0xFA;
					}
				}
				else // Flag Carry is 1
				{
					if ((x & HALFCARRY) == 0)
					{
						if ((highNibble >= 0x07) && (lowNibble <= 0x09)) // Line 12
						{
							a+=0xA0;
							carry = true;

						}
					}
					else // Flag HalfCarry is 1
					{
						if ((highNibble >= 0x06) && (lowNibble >= 0x06)) // Line 13
						{
							a +=0x9A;
							carry = true;
						}
					}
				}
			}

			a &= 0xFF;
			if (carry) a|=0x100;

			daa[x] = a;
		} // end for
	}
}

